#include "stetrahedron.h"

//////////////////////////////////////////////////////////////////////
// STetrahedron Class
//

STetrahedron::STetrahedron()
{
  Vector v1, v2, v3;
  FLOAT f, x2, y1, y2, z;

  m_fDihedralAngle = acos(1.0/3.0); // 70.52

  // Tile
  y1 = 1/sqrt(3);
  x2 = 0.25;            // 0.5 - (0.5*cos(DEGTORAD(60.0)));
  y2 = 1/(4.0*sqrt(3)); // (0.5*sin(DEGTORAD(60.0))) - (1/(2.0*sqrt(3)));
  z  = sqrt(6)/12.0;

  v1.Set(0.0, 0.0, z);
  v2.Set( x2,  y2, z);
  v3.Set(0.0,  y1, z);

  f = 1.0 / v3.Magnitude();

  v1 *= f;
  v2 *= f;
  v3 *= f;

  // 2D
  m_pt2DTilePoint1.Set(v1[X], v1[Y], v1[Z]);
  m_pt2DTilePoint2.Set(v2[X], v2[Y], v2[Z]);
  m_pt2DTilePoint3.Set(v3[X], v3[Y], v3[Z]);

  Init();
}

STetrahedron::~STetrahedron()
{
}

VOID STetrahedron::InitTileGrid()
{
  // 2D 
  m_pTile->InsertTileGridPoint( m_pt2DTilePoint1 );
  m_pTile->InsertTileGridPoint( m_pt2DTilePoint2 );
  m_pTile->InsertTileGridPoint( m_pt2DTilePoint3 );
}

VOID STetrahedron::InitFaceGrid()
{
  Vector v1, v2, v3;
  FLOAT y1, y2, z;

  y1 = 1/sqrt(3);
  y2 = -1/(2.0*sqrt(3));
  z  = sqrt(6)/12.0;

  v1.Set( 0.0, y1, z);
  v2.Set(-0.5, y2, z);
  v3.Set( 0.5, y2, z);

  v1.Normalize();
  v2.Normalize();
  v3.Normalize();

  // 2D
  m_pTile->InsertFaceGridPoint(v1[X], v1[Y], v1[Z]);
  m_pTile->InsertFaceGridPoint(v2[X], v2[Y], v2[Z]);
  m_pTile->InsertFaceGridPoint(v3[X], v3[Y], v3[Z]);
}

VOID STetrahedron::InitSphereTileGrid()
{
  // [pt1 -> pt2 -> pt3]
  AddSphereTileGridTesselatedPoints(m_pt2DTilePoint1, m_pt2DTilePoint2);
  AddSphereTileGridTesselatedPoints(m_pt2DTilePoint2, m_pt2DTilePoint3);
  AddSphereTileGridTesselatedPoints(m_pt2DTilePoint3, m_pt2DTilePoint1);
}

VOID STetrahedron::InitSphereFaceGrid()
{
  FLOAT y1, y2, z;
  Point pt, pt1, pt2, pt3;
  Vector v;

  // 3D
  y1 = -1/(2.0*sqrt(3));
  y2 = 1/sqrt(3);
  z  = sqrt(6)/12.0;

  pt1.Set( 0.0, y2, z);
  pt2.Set(-0.5, y1, z);
  pt3.Set( 0.5, y1, z);

  // [pt1 -> pt2 -> pt3]
  AddSphereFaceGridTesselatedPoints(pt1, pt2);
  AddSphereFaceGridTesselatedPoints(pt2, pt3);
  AddSphereFaceGridTesselatedPoints(pt3, pt1);
}

VOID STetrahedron::InitOutline()
{
  TilePoint *tpt1, *tpt2, *tpt3;

  // 2D
  tpt1 = m_pTile->InsertOutlinePoint(m_pt2DTilePoint1, tpTILE, 1 );
  tpt2 = m_pTile->InsertOutlinePoint(m_pt2DTilePoint2, tpTILE, 2 );
  tpt3 = m_pTile->InsertOutlinePoint(m_pt2DTilePoint3, tpTILE, 3 );

  tpt1->SetPartner(tpt1);
  tpt2->SetPartner(tpt2);
  tpt3->SetPartner(tpt3);

  m_pTile->TesselateOutlineAll();
}

VOID STetrahedron::InitFaceMatrices()
{
  MATRIX4D m;   
  Matrix *mm;

  m_listFaceTransformMatrix.RemoveAll();

  glMatrixMode(GL_MODELVIEW);

  glPushMatrix();
  {
    glLoadIdentity();
    
    // Face 1
    glPushMatrix();
    {
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    // Face 2
    glPushMatrix();
    {
      glTranslatef(0, 0, sqrt(6)/12.0); 
      glTranslatef(0, -1/(2.0*sqrt(3)), 0); 
      glRotatef( 180 - RADTODEG(m_fDihedralAngle), 1, 0, 0);
      glTranslatef(0, -1/(2.0*sqrt(3)), 0);
      glRotatef( 180, 0, 0, 1);
      glTranslatef(0, 0, -sqrt(6)/12.0); 
      
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    // Face 3
    glPushMatrix();
    {
      glTranslatef(0, 0, sqrt(6)/12.0); 
      glRotatef( 120, 0, 0, 1);
      glTranslatef(0, -1/(2.0*sqrt(3)), 0); 
      glRotatef( 180 - RADTODEG(m_fDihedralAngle), 1, 0, 0);
      glTranslatef(0, -1/(2.0*sqrt(3)), 0);
      glRotatef( 180, 0, 0, 1);
      glTranslatef(0, 0, -sqrt(6)/12.0); 
      
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    // Face 4
    glPushMatrix();
    {
      glTranslatef(0, 0, sqrt(6)/12.0); 
      glRotatef( -120, 0, 0, 1);
      glTranslatef(0, -1/(2.0*sqrt(3)), 0); 
      glRotatef( 180 - RADTODEG(m_fDihedralAngle), 1, 0, 0);
      glTranslatef(0, -1/(2.0*sqrt(3)), 0);
      glRotatef( 180, 0, 0, 1);
      glTranslatef(0, 0, -sqrt(6)/12.0); 
      
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();
    
  }
  glPopMatrix();
}

VOID STetrahedron::InitTileMatrices()
{
  Matrix *mm;
  Matrix m1, m2;
  Vector v;

  m_listTileTransformMatrix.RemoveAll();

  // Tile 1
  mm = new Matrix(); // Identity
  m_listTileTransformMatrix.Insert(mm, 0);

  // Tile 2
  mm = new Matrix();
  mm->SetRotateZ(DEGTORAD(120.0));
  m_listTileTransformMatrix.Insert(mm, 0);

  // Tile 3
  mm = new Matrix();
  mm->SetRotateZ(DEGTORAD(-120.0));
  m_listTileTransformMatrix.Insert(mm, 0);

  // Tile 4
  mm = new Matrix();
  v.Set(-1, 1, 1);
  mm->SetScale(v);
  m_listTileTransformMatrix.Insert(mm, 0);

  // Tile 5
  mm = new Matrix();
  v.Set(-1, 1, 1);
  m1.SetScale(v);
  m2.SetRotateZ(DEGTORAD(120.0));
  m1 = m2 * m1;
  mm->Set(m1.Get());
  m_listTileTransformMatrix.Insert(mm, 0);

  // Tile 6
  mm = new Matrix();
  v.Set(-1, 1, 1);
  m1.SetScale(v);
  m2.SetRotateZ(DEGTORAD(-120.0));
  m1 = m2 * m1;
  mm->Set(m1.Get());
  m_listTileTransformMatrix.Insert(mm, 0);
}

VOID STetrahedron::InsertOutlinePoint(UINT_32 uiSelectedPoint, Point pt)
{
  // Do nothing
}

VOID STetrahedron::FindMatrices(Matrix &m1, Matrix &m2, TilePoint *tpt1)
{
  Vector v;

  if( tpt1->GetType() != tpTILE )
    {
      m1 = tpt1->GetPartnerMatrix();

      if( tpt1->GetPartner() != NULL )
	m2 = (tpt1->GetPartner())->GetPartnerMatrix();
    }
  else if( tpt1->GetTilePoint() == 1 )
    {
      // No partner
    }
  else if( tpt1->GetTilePoint() == 2 )
    {
      // No partner
    }
  else if( tpt1->GetTilePoint() == 3 )
    {
      // No partner
    }
  else if( tpt1->GetTilePoint() == 4 )
    {
      // No partner
    }
  else
    {
      ASSERT(FALSE);
    }
}

///////////////////////////////////////////////////////////////////////////////

VOID STetrahedron::OutputES(FILE *fpOutputFile, CHAR *paletteFilename,
			    CHAR *colorFilename)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "stetrahedron tile 0\n\n" );

  m_pTile->OutputES(fpOutputFile);

  fprintf(fpOutputFile, "palettefile %s\n", paletteFilename);
  fprintf(fpOutputFile, "colorindexfile %s\n", colorFilename);
  //OutputESColor(fpOutputFile);
}
