#ifndef __TILE_H__
#define __TILE_H__

#include "mathlib.h"
#include "datalib.h"
#include "tricall.h"
#include "tilepoint.h"


class SymmetryGroup;

//////////////////////////////////////////////////////////////////////////////
// Tile
//

class Tile
{

//////////////////////////////////////////////////
// Initialization
//
public:
  Tile();
  ~Tile();

  inline VOID SetSGParent(SymmetryGroup *pSG);

  VOID InsertTileGridPoint(Point pt);
  VOID InsertFaceGridPoint(FLOAT x, FLOAT y, FLOAT z);
  VOID InsertSphereFaceGridPoint(Point pt);
  VOID InsertSphereTileGridPoint(Point pt);

  TilePoint *InsertOutlinePoint(Point pt, TilePointType tpType,	UINT_32 uiTilePoint);

  VOID CalculateIndexedPoints();
  TilePoint *GetIndexedPoint(UINT_32 index);

  UINT_32 GetPointIndex(TilePoint *tpt);

//////////////////////////////////////////////////
// Rendering
//
public:
  VOID PrintVertexFacePoints();
  VOID PrintVertexTilePoints();

  VOID RenderSphereFace();
  VOID RenderSphereTile();
  VOID RenderSphereOutlineEdges();
  VOID RenderSphereRaisedOutlineEdges();
  VOID RenderSphereOutlinePoints();
  VOID RenderSphereInnerEdges();
  VOID RenderSphereInnerPoints();
  VOID RenderSphereHolePoints();

  VOID RenderHeightOutlineEdges();
  VOID RenderHeightOutlinePoints();
  VOID RenderHeightOutlinePointNormals();
  VOID RenderHeightInnerEdges();
  VOID RenderHeightInnerPoints();
  VOID RenderHeightInnerPointNormals();

  VOID RenderSelectionSphereOutlinePoints();
  VOID RenderSelectionSphereOutlineEdges();
  VOID RenderSelectionSphereInnerPoints();
  VOID RenderSelectionSphereInnerEdges();
  VOID RenderSelectionSphereHeightOutlinePoints();
  VOID RenderSelectionSphereHeightInnerPoints();
  VOID RenderSelectionSphereHolePoints();

  VOID RenderMesh();                   // stereographic
  VOID RenderStereographicPoints();
  VOID RenderStereographicEdges();
  VOID RenderMeshTilePoints();         // triangulated tile 
  VOID RenderMeshSide(Point pt1, Point pt2, Point pp1, Point pp2);
  VOID RenderMeshExtrudedTilePoints(BOOL bBorder); // solid

//////////////////////////////////////////////////
// Output Functions
//
public:
  VOID OutputES(FILE *fpOutputFile);
  VOID OutputESTilePointPartners(FILE *fpOutputFile);

  VOID OutputSTL(FILE *fpOutputFile, Matrix mTransform, BOOL bBorder);
  VOID OutputVRML(FILE *fpOutputFile, Matrix mTransform, Vector vColor, BOOL bBorder);

  VOID ClearSphereOutlinePoints();
  VOID ClearSphereInnerPoints();
  VOID ClearSphereHolePoints();
  VOID CreateTilePoint(Point pt, UINT_32 uiType, UINT_32 uiTilePoint, 
		       UINT_32 uiConnected, FLOAT fHeight, Matrix m);
  VOID SetPartners(UINT_32 uiID, UINT_32 uiIDPartner);
  VOID SetHeightPartnersFromOpenES(UINT_32 uiID, UINT_32 uiIDPartner);

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  inline VOID SetSolidModified(BOOL b);
  inline BOOL GetSolidModified();

  inline VOID SetBasReliefModified(BOOL b);
  inline BOOL GetBasReliefModified();

  inline VOID SetRadius(FLOAT fRadius);
  inline VOID SetThickness(FLOAT fThickness);

  // Stereographic Projection
  inline Point GetNorthPole();
  Point StereographicProjection(Point pt);

  // Points
  UINT_32 FindTileSide(UINT_32 uiSelectedPoint);
  VOID MoveOutlinePoint(UINT_32 uiSelectedPoint, Point ptLocation);
  VOID InsertOutlinePoint(UINT_32 uiSelectedPoint, Point ptLocation);
  VOID DeleteOutlinePoint(UINT_32 uiSelectedPoint);
  VOID EnumerateOutlinePoints();

  TilePoint *AddInnerPoint(Point ptLocation);
  TilePoint *AddInnerPointCopy(UINT_32 uiSelectedType, UINT_32 uiSelectedPoint);
  VOID MoveInnerPoint(UINT_32 uiSelectedPoint, Point ptLocation);
  VOID InsertInnerPoint(UINT_32 uiSelectedPoint, Point ptLocation);
  VOID DeleteInnerPoint(UINT_32 uiSelectedPoint);
  VOID SnapInnerPointToSelected(UINT_32 uiInnerPointIndex, UINT_32 uiSelectedType,
				UINT_32 uiSelectedIndex);
  VOID ClearSnapPoint(UINT_32 uiInnerPointIndex);

  VOID AddHolePoint(Point ptLocation);
  VOID MoveHolePoint(UINT_32 uiSelectedPoint, Point ptLocation);
  VOID DeleteHolePoint(UINT_32 uiSelectedPoint);

  BOOL GetPointHeightSelected(UINT_32 uiSelectedType, UINT_32 uiSelectedPoint);
  VOID SetPointHeightSelected(UINT_32 uiSelectedType, UINT_32 uiSelectedPoint, BOOL b);
  VOID ToglePointHeightSelected(UINT_32 uiSelectedType, UINT_32 uiSelectedPoint);
  VOID DeselectAllHeightPoints();
  VOID AddHeightToPoints(FLOAT fHeight);
  Vector GetPointNormalVector(UINT_32 uiSelectedPoint);

  VOID Tesselate(TilePoint *tpt1, TilePoint *tpt2);

  VOID TesselateOutlineAll();
  VOID TesselateOutline(TilePoint *tpt);
  VOID TesselateOutlineNeighbors(TilePoint *tpt);

  VOID TesselateInnerAll();
  VOID TesselateInner(TilePoint *tpt);
  VOID TesselateInnerBefore(TilePoint *tpt);
  VOID TesselateInnerNeighbors(TilePoint *tpt);

//////////////////////////////////////////////////
// Functions
//
  VOID CalculateNorthPole();
  VOID CalculateStereographicPoints();
  VOID Triangulate();
  VOID TriangulateByFile();
  VOID TriangulateByAPI();
  VOID Foo();

  VOID ClearMesh(struct triangulateio &mesh);
  VOID InitMesh(struct triangulateio &mesh);

  BOOL SphereOutlinePoint(TilePoint *tpt);
  VOID CalculateExtrudedPoints(Point &p1, Point &pp1, TilePoint *);
  VOID CalculateOutputExtrudedPoints(Point &p1, Point &pp1, TilePoint *);
  VOID CalculateHoleEdges();

//////////////////////////////////////////////////
// Class Variables
//
protected:

//////////////////////////////////////////////////
// Member Variables
//
protected:

  SymmetryGroup *m_pSGParent;
  BOOL m_bSolidModified;
  BOOL m_bBasReliefModified;

  UINT_32 m_uiPoints;
  TilePoint **m_ppIndexedTilePoints;

  UINT_32 m_uiHoleEdges;

  // Stereographic Projection NorthPole
  Point m_ptNorthPole;
  struct triangulateio triMesh;
  struct triangulateio in;

  // Tile
  FLOAT m_fRadius;
  FLOAT m_fThickness;

  // Lists of points
  CListInt m_listFaceGridPoints;        // Grid for polyhedra face
  CListInt m_listTileGridPoints;        // Grid for tile

  CListInt m_listSphereFaceGridPoints;  // Face Grid projected on sphere
  CListInt m_listSphereTileGridPoints;  // Tile Grid projected on sphere

  CListInt m_listSphereOutlinePoints;   // Outline pts projected on sphere
  CListInt m_listSphereInnerPoints;     // Inner pts projected on sphere

  // List of holes
  CListInt m_listSphereHolePoints;
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline VOID Tile::SetSolidModified(BOOL b)
{
  m_bSolidModified = b;
}

inline BOOL Tile::GetSolidModified()
{
  return m_bSolidModified;
}

inline VOID Tile::SetBasReliefModified(BOOL b)
{
  m_bBasReliefModified = b;
}

inline BOOL Tile::GetBasReliefModified()
{
  return m_bBasReliefModified;
}

inline VOID Tile::SetRadius(FLOAT fRadius)
{
  m_fRadius = fRadius;
}

inline VOID Tile::SetThickness(FLOAT fThickness)
{
  m_fThickness = fThickness;

  m_bSolidModified = TRUE;
}

inline VOID Tile::SetSGParent(SymmetryGroup *pSG)
{
  m_pSGParent = pSG;
}

inline Point Tile::GetNorthPole()
{
  return m_ptNorthPole;
}

#endif // __TILE_H__
