#ifndef __TILEPOINT_H__
#define __TILEPOINT_H__

#include "mathlib.h"
#include "datalib.h"
#include "snappoint.h"


enum TilePointType 
{
  tpTILE,
  tpOUTLINE,
  tpINNER,
  tpHOLE,
  tpSize
};

//////////////////////////////////////////////////////////////////////////////
// TilePoint
//

class TilePoint
{

//////////////////////////////////////////////////
// Initialization
//
public:
  TilePoint();
  virtual ~TilePoint();

  VOID RemoveAllTesselatedPoints();
  VOID InsertTesselatedPoint(Point pt);

//////////////////////////////////////////////////
// Rendering
//
public:
  VOID PrintVertexTesselatedPoints();
  VOID PrintVertexRaisedTesselatedPoints();

//////////////////////////////////////////////////
// Output
//
public:
  VOID OutputES(FILE *fpOutputFile);

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  inline Point GetPoint();
  inline VOID SetPoint(Point pt);
  inline VOID SetPoint(FLOAT x, FLOAT y, FLOAT z, FLOAT w = 1.0f);

  inline VOID SetHeightPoint(Point pt);
  inline Point GetHeightPoint();

  inline VOID SetHeight(FLOAT f);
  inline VOID SetHeightValue(FLOAT f);
  inline FLOAT GetHeightValue();

  inline Point GetStereographicPoint();
  inline VOID SetStereographicPoint(Point pt);

  inline TilePointType GetType();
  inline VOID SetType(TilePointType tpType);

  inline UINT_32 GetTilePoint();
  inline VOID SetTilePoint(UINT_32 uiTilePoint);

  inline UINT_32 GetID();
  inline VOID SetID(UINT_32 uiID);

  inline TilePoint *GetPartner();
  inline VOID SetPartner(TilePoint *ptp);

  inline Matrix GetPartnerMatrix();
  inline VOID SetPartnerMatrix(Matrix m);

  inline TilePoint* GetHeightPartner();
  inline VOID SetHeightPartner(TilePoint *tpt);
  inline VOID SetHeightPartnerFromOpenES(TilePoint *tpt);

  inline BOOL GetConnected();
  inline VOID SetConnected(BOOL bConnected);

  inline BOOL GetHeightSelected();
  inline VOID SetHeightSelected(BOOL bHeightSelected);

  inline VOID SetSnapPointSnapped(BOOL b);

  inline SnapPoint* GetSnapPoint();
  inline VOID SetSnapPoint(SnapPoint *psp);
  VOID ClearSnapPoint();
  VOID DeleteSnapPoint();

  VOID CalculateHeightPoint();
  VOID AddHeight(FLOAT fHeight);
  VOID SnapToHeight(FLOAT fHeight, Point pt);

  VOID CalculateCentroid(Point &ptCentroid, UINT_32 &uiPoints);

//////////////////////////////////////////////////
// Class Variables
//
protected:

//////////////////////////////////////////////////
// Member Variables
//
protected:

  UINT_32 m_uiID;           // Points enumerated

  TilePointType m_tpType;   // tile, outline, inner
  UINT_32 m_uiTilePoint;    // Which edge this point is on

  Point m_pt;               // Position
  Point m_ptStereographic;  // Stereographic projected point
  FLOAT m_fHeight;          // Height along normal
  Point m_ptHeight;         // Point along normal

  CListInt m_listTesselatedPoint; // Tesselated pts eminating from this tile point

  // For outline points
  TilePoint *m_ptpPartner;       // symmetry partner
  Matrix m_mPartnerTransform;    // matrix to apply to this pt to get partner pt
  TilePoint *m_ptpHeightPartner; // height symmetry partner

  // For inner points
  BOOL m_bConnected;             // mark if connected to next pt

  // For assigning height to inner points
  SnapPoint *m_pSnapPoint;
  BOOL m_bHeightSelected;
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline Point TilePoint::GetPoint()
{
  return m_pt;
}

inline VOID TilePoint::SetPoint(Point pt)
{
  m_pt = pt;

  CalculateHeightPoint();
}

inline VOID TilePoint::SetPoint(FLOAT x, FLOAT y, FLOAT z, FLOAT w/* = 1.0f*/)
{
  m_pt.Set(x, y, z, w);

  CalculateHeightPoint();
}

inline VOID TilePoint::SetHeightPoint(Point pt)
{
  m_ptHeight = pt;
}

inline Point TilePoint::GetHeightPoint()
{
  return m_ptHeight;
}

inline VOID TilePoint::SetHeightValue(FLOAT f)
{
  m_fHeight = f;
}

inline FLOAT TilePoint::GetHeightValue()
{
  return m_fHeight;
}

inline VOID TilePoint::SetHeight(FLOAT f)
{
  m_fHeight = f;
  CalculateHeightPoint();
}

inline Point TilePoint::GetStereographicPoint()
{
  return m_ptStereographic;
}

inline VOID TilePoint::SetStereographicPoint(Point pt)
{
  m_ptStereographic = pt;
}

inline TilePointType TilePoint::GetType()
{
  return m_tpType;
}

inline VOID TilePoint::SetType(TilePointType tpType)
{
  m_tpType = tpType;
}

inline UINT_32 TilePoint::GetTilePoint()
{
  return m_uiTilePoint;
}

inline VOID TilePoint::SetTilePoint(UINT_32 uiTilePoint)
{
  m_uiTilePoint = uiTilePoint;
}

inline UINT_32 TilePoint::GetID()
{
  return m_uiID;
}

inline VOID TilePoint::SetID(UINT_32 uiID)
{
  m_uiID = uiID;
}

inline TilePoint *TilePoint::GetPartner()
{
  return m_ptpPartner;
}

inline VOID TilePoint::SetPartner(TilePoint *ptp)
{
  m_ptpPartner = ptp;
}

inline Matrix TilePoint::GetPartnerMatrix()
{
  return m_mPartnerTransform;
}

inline VOID TilePoint::SetPartnerMatrix(Matrix m)
{
  m_mPartnerTransform = m;
}

inline TilePoint* TilePoint::GetHeightPartner()
{
  return m_ptpHeightPartner;
}

inline VOID TilePoint::SetHeightPartner(TilePoint *tpt)
{
  m_ptpHeightPartner = tpt;
}

inline VOID TilePoint::SetHeightPartnerFromOpenES(TilePoint *tpt)
{
  if( m_tpType != tpTILE )
    m_ptpHeightPartner = tpt;
}

inline BOOL TilePoint::GetConnected()
{
  return m_bConnected;
}

inline VOID TilePoint::SetConnected(BOOL bConnected)
{
  m_bConnected = bConnected;
}

inline BOOL TilePoint::GetHeightSelected()
{
  return m_bHeightSelected;
}

inline VOID TilePoint::SetHeightSelected(BOOL bHeightSelected)
{
  m_bHeightSelected = bHeightSelected;
}

inline VOID TilePoint::SetSnapPointSnapped(BOOL b)
{
  if( m_pSnapPoint != NULL )
    {
      m_pSnapPoint->SetSnapped(b);
    }
}

inline SnapPoint* TilePoint::GetSnapPoint()
{
  return m_pSnapPoint;
}

inline VOID TilePoint::SetSnapPoint(SnapPoint* psp)
{
  m_pSnapPoint = psp;
}

#endif // __TILEPOINT_H__
