#ifndef __HASHTABLE_H__
#define __HASHTABLE_H__

#include "genlib.h"

//////////////////////////////////////////////////////////////////////////////
// C++ Basic type List Functions
//

inline UINT_32 Hash(int i) { return (UINT_32)i; }
inline UINT_32 Hash(INT_32 i) { return (UINT_32)i; }
inline UINT_32 Hash(UINT_32 i) { return i; }

//////////////////////////////////////////////////////////////////////////////
// C++ Basic type List Functions
//

template <class TList, class TPData, class TKey>
class CHashTable
{
public:
  CHashTable();
  ~CHashTable();
  BOOL Init(INT_32 iSize = 127);

protected:
  VOID Uninit();

public:
  TPData Insert(TPData pData, TKey key); //inserts entry into HashTable
  TPData Peek(TKey key);
  TPData Remove(TKey key);
  TPData RemoveFirst();
  VOID Display();			//displays the HashTable

protected:
  VOID Rehash();			//rehashes HashTable
  
protected:
  TList *m_plistBuckets;
  UINT_32 m_uiBuckets;
  UINT_32 m_uiSize;
  UINT_32 m_uiVolume;  
};

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
// Class Implementations
//

inline BOOL IsPrime(UINT_32 n)
//Requires:  int n > 0.
//Effect:  returns TRUE if c is a prime number.
{
  UINT_32 i;

  // Check if even
  if ( !(n & 0x1) )
    {
      return FALSE;
    }
  
  for ( i=3; i*i <= n; i += 2 )
    {
      if ( n % i == 0 )
        {
          return FALSE;
        }
    }
  return TRUE;
}

inline UINT_32 GetNextPrime(UINT_32 n)
//Requires:  int n > 0.
//Effect:  returns the next prime number >= n.
{
  if ( !(n & 0x1) )
    {
      n++;
    }

  while ( !IsPrime(n) )
    {
      n+=2;
    }

  return n;
}

//////////////////////////////////////////////////////////////////////////////
// CHashTable Class
//

template <class TList, class TPData, class TKey>
CHashTable<TList, TPData, TKey>::CHashTable()
{
  m_uiVolume = 0;
  m_uiSize = 0;
  m_uiBuckets = 0;
  m_plistBuckets = NULL;
}

template <class TList, class TPData, class TKey>
CHashTable<TList, TPData, TKey>::~CHashTable()
{
  Uninit();
}

template <class TList, class TPData, class TKey>
BOOL CHashTable<TList, TPData, TKey>::Init(INT_32 uiBuckets/* = 127*/)
{
  ASSERT( m_plistBuckets == NULL && m_uiBuckets == 0 ); 

  UINT_32 i;

  m_uiVolume = 0;
  m_uiSize = 0;
  m_uiBuckets = GetNextPrime(uiBuckets);

  m_plistBuckets = new TList[m_uiBuckets];
  for ( i = 0; i < m_uiBuckets; i++ )
    {
      m_plistBuckets[i].Init();
    }

  return TRUE;
}

template <class TList, class TPData, class TKey>
VOID CHashTable<TList, TPData, TKey>::Uninit()
{
  if ( m_plistBuckets != NULL )
    {
      delete []m_plistBuckets;
      m_plistBuckets = NULL;
    }
}

template <class TList, class TPData, class TKey>
TPData CHashTable<TList, TPData, TKey>::Insert(TPData pData, TKey key)
{
  ASSERT( m_plistBuckets != NULL && m_uiBuckets > 0 ); 

  TPData pResult;
  UINT_32 h;

  h = ::Hash(key) % m_uiBuckets;

  pResult = m_plistBuckets[h].Insert(pData, key);

  if ( pResult == pData )
    {
      m_uiVolume++;

      if ( m_uiVolume >= (3*m_uiBuckets/2) )
	{
	  Rehash();
	}
    }
  // otherwise it was a duplicate and we ignore it

  return pResult;
}

template <class TList, class TPData, class TKey>
void CHashTable<TList, TPData, TKey>::Rehash()
{
  ASSERT( m_plistBuckets != NULL && m_uiBuckets > 0 ); 

  TList *plist;
  UINT_32 i, uiBuckets;
  TPData pData;
  TKey key;

  //printf("\nRehash!!!\n");

  plist = m_plistBuckets;
  uiBuckets = m_uiBuckets;
  
  m_uiVolume = 0;
  m_uiSize = 0;
  m_uiBuckets = GetNextPrime(2*m_uiBuckets);
  m_plistBuckets = new TList[m_uiBuckets];
  for ( i = 0; i < m_uiBuckets; i++ )
    {
      m_plistBuckets[i].Init();
    }
  
  for ( i = 0; i < uiBuckets; i++ )
    {
      while( (pData = plist[i].RemoveFirst(&key)) != NULL )
	{
	  Insert(pData, key);
	}
    }

  delete []plist;
}

template <class TList, class TPData, class TKey>
TPData CHashTable<TList, TPData, TKey>::Peek(TKey key)
{
  ASSERT( m_plistBuckets != NULL && m_uiBuckets > 0 ); 

  TPData pResult;
  UINT_32 h;

  h = ::Hash(key) % m_uiBuckets;

  pResult = m_plistBuckets[h].Peek(key);
 
  return pResult;
}

template <class TList, class TPData, class TKey>
TPData CHashTable<TList, TPData, TKey>::Remove(TKey key)
{
  ASSERT( m_plistBuckets != NULL && m_uiBuckets > 0 ); 

  TPData pResult;
  UINT_32 h;

  h = ::Hash(key) % m_uiBuckets;

  pResult = m_plistBuckets[h].Remove(key);
 
  return pResult;
}

template <class TList, class TPData, class TKey>
TPData CHashTable<TList, TPData, TKey>::RemoveFirst()
{
  ASSERT( m_plistBuckets != NULL && m_uiBuckets > 0 ); 

  TPData pResult;
  UINT_32 h;

  for ( h = 0; h < m_uiBuckets; h++ )
    {
      pResult = m_plistBuckets[h].RemoveFirst();
      if ( pResult != NULL )
	{
	  break;
	}
    }

  return pResult;
}

template <class TList, class TPData, class TKey>
VOID CHashTable<TList, TPData, TKey>::Display()
{
  ASSERT( m_plistBuckets != NULL && m_uiBuckets > 0 ); 

  UINT_32 i;

  printf("\n");
  for ( i = 0; i < m_uiBuckets; i++)
    {
      printf("[%ld]->", i);
      m_plistBuckets[i].Display();
    }
}

#endif // __HASHTABLE_H__

