#include "tile.h"
#include "stl.h"
#include "symmetrygroup.h"
#include "main.h"

extern BOOL gbCreateTriFiles;
extern BOOL gbExtrudeInward;
extern BOOL gbTileDetailed;
extern BOOL gbZoomIn;
extern GLuint GLuiSphere02_8;
extern GLuint GLuiSphere01_8;
extern GLuint GLuiSphere005_8;
extern VOID SetStatusLabel(CHAR *pcLabel);
extern UINT_32 guiCurrentSelectedInner;
extern CHAR * gpcTriangleCommand;
extern CHAR ** gppcTriangleArgs;

/*
extern "C" {
  extern void triangulate(char *, struct triangulateio *, struct triangulateio *,
			  struct triangulateio *);
}
*/

extern "C" {
  extern int mainTri(int argc, char** argv);
}

VOID safefree(VOID* p)
{
  if( NULL != p )
    {
      free(p);
    }
}

//////////////////////////////////////////////////////////////////////
// Tile Class
//

Tile::Tile()
{
  m_bSolidModified = FALSE;
  m_bBasReliefModified = FALSE;

  m_fRadius = 1.0;
  m_fThickness = 0.02;

  m_ppIndexedTilePoints = NULL;
  m_pSGParent = NULL;

  m_uiHoleEdges = 0;

  m_uiPoints = 0;

  InitMesh(triMesh);
  InitMesh(in);

  m_listFaceGridPoints.Init();
  m_listTileGridPoints.Init();

  m_listSphereFaceGridPoints.Init();
  m_listSphereTileGridPoints.Init();
  m_listSphereOutlinePoints.Init();
  m_listSphereInnerPoints.Init();

  m_listSphereHolePoints.Init();
}

Tile::~Tile()
{
  safefree(triMesh.pointlist);
  safefree(triMesh.pointattributelist);
  safefree(triMesh.pointmarkerlist);
  safefree(triMesh.trianglelist);
  safefree(triMesh.triangleattributelist);
  safefree(triMesh.trianglearealist);
  safefree(triMesh.neighborlist);
  safefree(triMesh.segmentlist);
  safefree(triMesh.segmentmarkerlist);
  safefree(triMesh.holelist);
  safefree(triMesh.regionlist);
  safefree(triMesh.edgelist);
  safefree(triMesh.edgemarkerlist);
  safefree(triMesh.normlist);
}

VOID Tile::CalculateIndexedPoints()
{
  UINT_32 i, uiPoints, uiOuterPoints, uiInnerPoints;
  CListIterInt iter;
  TilePoint *tpt;

  if( m_ppIndexedTilePoints != NULL )
    {
      delete m_ppIndexedTilePoints;
    }

  uiOuterPoints = m_listSphereOutlinePoints.GetLength();
  uiInnerPoints = m_listSphereInnerPoints.GetLength();
  uiPoints = uiOuterPoints + uiInnerPoints;

  m_uiPoints = uiPoints;
  m_ppIndexedTilePoints = new TilePoint*[uiPoints];

  iter.Init(&m_listSphereOutlinePoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()), i=0; 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()), i++ )
    {
      ASSERT( i < uiPoints );
      m_ppIndexedTilePoints[i] = tpt;
    }
  ASSERT( i == uiOuterPoints );

  iter.Init(&m_listSphereInnerPoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()); 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()), i++ )
    {
      ASSERT( i < uiPoints );
      m_ppIndexedTilePoints[i] = tpt;
    }
  ASSERT( i == uiPoints );
}

TilePoint* Tile::GetIndexedPoint(UINT_32 index)
{
  //ASSERT( index < m_uiPoints );

  if( index  >= m_uiPoints )
    {
      printf( "Triangulation failed because of bad outer contour.\n");
      printf( "Please uncross all edges.\n");
      return NULL;
    }

  return (m_ppIndexedTilePoints[index]);
}

UINT_32 Tile::GetPointIndex(TilePoint *tpt)
{
  UINT_32 i;

  for( i=0; i < m_uiPoints; i++ )
    {
      if( tpt == m_ppIndexedTilePoints[i] )
	{
	  return i;
	}
    }

  ASSERT(FALSE);
  return 0;
}

VOID Tile::CalculateHoleEdges()
{
  UINT_32 i, index;
  TilePoint *tpt1;

  m_uiHoleEdges = 0;

  if( triMesh.edgemarkerlist != NULL )
    {
      for( i = 0; i < (UINT_32)triMesh.numberofedges; i++ )
	{
	  if( 1 == triMesh.edgemarkerlist[i] ) // Boundary
	    {
	      index = triMesh.edgelist[2*i];
	      tpt1 = GetIndexedPoint(index);
		  
	      if( !SphereOutlinePoint(tpt1) ) // Hole
		{
		  m_uiHoleEdges++;
		}	  
	    }
	}
    }
  //printf( "HoleEdges: %lu\n", m_uiHoleEdges );
}

VOID Tile::InsertSphereFaceGridPoint(Point p)
{
  Point *pt;

  pt = new Point();
  pt->Set(p.Get());

  m_listSphereFaceGridPoints.InsertLast(pt, 0);
}

VOID Tile::InsertSphereTileGridPoint(Point p)
{
  Point *pt;

  pt = new Point();
  pt->Set(p.Get());

  m_listSphereTileGridPoints.InsertLast(pt, 0);
}

VOID Tile::InsertFaceGridPoint(FLOAT x, FLOAT y, FLOAT z)
{
  Point *pt;

  pt = new Point();
  pt->Set(x, y, z);

  m_listFaceGridPoints.InsertLast(pt, 0);
}

VOID Tile::InsertTileGridPoint(Point p)
{
  Point *pt;

  pt = new Point();
  pt->Set(p.Get());

  m_listTileGridPoints.InsertLast(pt, 0);
}

/////////////////////////////////////////////////////////////////////////////////////

VOID Tile::Tesselate(TilePoint *tpt1, TilePoint *tpt2)
{
  UINT_32 i;
  FLOAT fSlices;
  Point pt, pt1, pt2;
  Vector v;

  pt1 = tpt1->GetPoint();
  pt2 = tpt2->GetPoint();
  fSlices = ceil( 50.0*(pt2-pt1).Magnitude() );
  fSlices = MAX( 1, fSlices);

  tpt1->RemoveAllTesselatedPoints();

  // [pt1 -> pt2)
  for( i = 0; i < fSlices; i++ )
    {
      pt = ((FLOAT)i/fSlices)*pt2 + (1.0 - ((FLOAT)i/fSlices))*pt1;
      v = pt;
      v[W] = 0.0;
      v.Normalize();
      pt.Set(v.Get());
      
      tpt1->InsertTesselatedPoint(pt);
    }
}

VOID Tile::TesselateOutlineAll()
{
  CListIterInt iter;
  TilePoint *tpt1, *tpt2, *tptFirst;

  iter.Init(&m_listSphereOutlinePoints);
  tpt1 = (TilePoint *)(iter.PeekFirst());
  if( tpt1 == NULL )
    {
      return;
    }

  tptFirst = tpt1;

  for( tpt2 = (TilePoint *)(iter.PeekNext());
       tpt2 != NULL;
       tpt2 = (TilePoint *)(iter.PeekNext()) )
    {
      Tesselate( tpt1, tpt2);

      tpt1 = tpt2;
    }

  Tesselate( tpt1, tptFirst);
}

VOID Tile::TesselateOutlineNeighbors(TilePoint *tpt)
{
  ASSERT( tpt->GetType() != tpTILE );

  CListIterInt iter;
  TilePoint *tpt1, *tpt2, *tpt3, *tptFirst;

  iter.Init(&m_listSphereOutlinePoints);
  tptFirst = (TilePoint *)(iter.PeekFirst());
  if( tptFirst == NULL )
    {
      return;
    }

  tpt1 = tptFirst;
  tpt2 = tptFirst;
  tpt3 = NULL;

  // Find the selected point and its neighbors
  for( tpt2 = (TilePoint *)(iter.PeekNext());
       tpt2 != NULL;
       tpt2 = (TilePoint *)(iter.PeekNext()) )
    {
      if( tpt2 == tpt )
	{
	  tpt3 = (TilePoint *)(iter.PeekNext());

	  if( tpt3 == NULL )
	    {
	      tpt3 = tptFirst;
	    }

	  break;
	}

      tpt1 = tpt2;
    }
  
  ASSERT( (tpt1 != NULL) && (tpt2 != NULL) && (tpt3 != NULL) );

  Tesselate( tpt1, tpt2);  // [pt1 -> pt2) 
  Tesselate( tpt2, tpt3);  // [pt2 -> pt3)
}

VOID Tile::TesselateOutline(TilePoint *tpt)
{
  CListIterInt iter;
  TilePoint *tpt1, *tpt2, *tptFirst;
  
  tpt1 = NULL;
  tpt2 = NULL;

  iter.Init(&m_listSphereOutlinePoints);
  tptFirst = (TilePoint *)(iter.PeekFirst());

  // Find the selected point and the point after it
  for( tpt1 = (TilePoint *)(iter.PeekFirst());
       tpt1 != NULL;
       tpt1 = (TilePoint *)(iter.PeekNext()) )
    {
      if( tpt1 == tpt )
	{
	  tpt2 = (TilePoint *)(iter.PeekNext());

	  if( tpt2 == NULL )
	    {
	      tpt2 = tptFirst;
	    }

	  break;
	}
    }

  ASSERT( (tpt1 != NULL) && (tpt2 != NULL) );

  Tesselate( tpt1, tpt2 );
}

VOID Tile::TesselateInnerAll()
{
  CListIterInt iter;
  TilePoint *tpt1, *tpt2;
  
  iter.Init(&m_listSphereInnerPoints);
  tpt1 = (TilePoint *)(iter.PeekFirst());
  if( tpt1 == NULL )
    {
      return;
    }
  
  for( tpt2 = (TilePoint *)(iter.PeekNext());
       tpt2 != NULL;
       tpt2 = (TilePoint *)(iter.PeekNext()) )
    {
      if( tpt1->GetConnected() )
	{
	  Tesselate( tpt1, tpt2 );
	}

      tpt1 = tpt2;
    }
}

VOID Tile::TesselateInner(TilePoint *tpt)
{
  CListIterInt iter;
  TilePoint *tpt1, *tpt2;
  
  tpt1 = NULL;
  tpt2 = NULL;

  // Find the selected point and the point after it
  iter.Init(&m_listSphereInnerPoints);
  for( tpt1 = (TilePoint *)(iter.PeekFirst());
       tpt1 != NULL;
       tpt1 = (TilePoint *)(iter.PeekNext()) )
    {
      if( tpt1 == tpt )
	{
	  tpt2 = (TilePoint *)(iter.PeekNext());
	  break;
	}
    }

  ASSERT( tpt1 != NULL );

  tpt1->RemoveAllTesselatedPoints();

  if( (tpt2 == NULL) || !tpt1->GetConnected() ) // Only one point
    return; 

  Tesselate( tpt1, tpt2 );
}

VOID Tile::TesselateInnerBefore(TilePoint *tpt)
{
  CListIterInt iter;
  TilePoint *tpt1, *tpt2;
  
  tpt1 = NULL;
  tpt2 = NULL;

  // Find the selected point and the point before it
  iter.Init(&m_listSphereInnerPoints);
  for( tpt2 = (TilePoint *)(iter.PeekFirst());
       tpt2 != NULL;
       tpt2 = (TilePoint *)(iter.PeekNext()) )
    {
      if( tpt2 == tpt )
	{
	  break;
	}

      tpt1 = tpt2;
    }

  ASSERT( tpt2 != NULL );

  if( tpt1 == NULL ) // Only one point
    return; 

  tpt1->RemoveAllTesselatedPoints();

  if( !tpt1->GetConnected() )
    return;

  Tesselate( tpt1, tpt2 );
}

VOID Tile::TesselateInnerNeighbors(TilePoint *tpt)
{
  ASSERT( tpt->GetType() != tpTILE );

  CListIterInt iter;
  TilePoint *tpt1, *tpt2, *tpt3;

  tpt1 = NULL;
  tpt2 = NULL;
  tpt3 = NULL;

  // Find the selected point and its neighbors
  iter.Init(&m_listSphereInnerPoints);
  for( tpt2 = (TilePoint *)(iter.PeekFirst());
       tpt2 != NULL;
       tpt2 = (TilePoint *)(iter.PeekNext()) )
    {
      if( tpt2 == tpt )
	{
	  tpt3 = (TilePoint *)(iter.PeekNext());

	  break;
	}

      tpt1 = tpt2;
    }

  ASSERT( tpt2 != NULL );

  // [pt1 -> pt2)
  if( (tpt1 != NULL) && (tpt1->GetConnected()) )
    {
      Tesselate( tpt1, tpt2 );
    }

  // [pt2 -> pt3)
  if( (tpt3 != NULL) && (tpt2->GetConnected()) )
    {
      Tesselate( tpt2, tpt3 );
    }
}

UINT_32 Tile::FindTileSide(UINT_32 uiSelectedPoint)
{
  UINT_32 i, uiTilePoint;
  CListIterInt iter;
  TilePoint *tpt;

  iter.Init(&m_listSphereOutlinePoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()), i=0; 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()), i++ )
    {
      if( tpt->GetType() == tpTILE )
	{
	  uiTilePoint = tpt->GetTilePoint();
	}

      if( i == uiSelectedPoint )
	{
	  return uiTilePoint;
	}
    }

  ASSERT( FALSE );
  
  return uiTilePoint;
}

///////////////////////////////////////////////////////////////////////////////////////

TilePoint *Tile::InsertOutlinePoint(Point pt, TilePointType tpType, UINT_32 uiTilePoint)
{
  TilePoint *tpt;
  Vector v;

  // 3D point
  tpt = new TilePoint();
  v.Set(pt.Get());
  v[W] = 0.0;
  v.Normalize();
  tpt->SetPoint(v[X], v[Y], v[Z]);
  tpt->SetType(tpType);
  tpt->SetTilePoint(uiTilePoint);

  m_listSphereOutlinePoints.InsertLast(tpt, (UINT_32)tpt);

  m_bSolidModified = TRUE;

  return tpt;
}

VOID Tile::MoveOutlinePoint(UINT_32 uiSelectedPoint, Point pt)
{
  UINT_32 i;
  CListIterInt iter;
  TilePoint *tpt, *tptPartner, *tptPrev, *tptPost;
  Matrix m;

  tptPrev = NULL;
  tptPost = NULL;

  iter.Init(&m_listSphereOutlinePoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()), i=0; 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()), i++ )
    {
      if( i == uiSelectedPoint )
	{
	  if( tpt->GetType() == tpTILE )
	    {
	      SetStatusLabel( "Yellow points cannot be moved" );
	      return;
	    }

	  tpt->SetPoint(pt);
	  TesselateOutlineNeighbors(tpt);

	  // unsnap
	  tpt->ClearSnapPoint();

	  // partner
	  tptPartner = tpt->GetPartner();
	  if( tptPartner != NULL )
	    {
	      m = tpt->GetPartnerMatrix();
	      pt = m * pt;
	      tptPartner->SetPoint(pt);
	      TesselateOutlineNeighbors(tptPartner);
	    }

	  break;
	}

      tptPrev = tpt;
    }

  m_bSolidModified = TRUE;
}

VOID Tile::InsertOutlinePoint(UINT_32 uiSelectedPoint, Point ptL)
{
  UINT_32 i;
  CListIterInt iter;
  TilePoint *tpt, *tptPartner, *tpt1, *tptFirst, *tptNew, *tptNewPartner;
  Matrix m1, m2;

  tpt1 = NULL;

  // New Point
  tptNew = new TilePoint();
  tptNew->SetPoint(ptL);
  tptNew->SetType(tpOUTLINE);

  iter.Init(&m_listSphereOutlinePoints);
  tptFirst = (TilePoint *)(iter.PeekFirst());

  for( tpt1 = (TilePoint *)(iter.PeekFirst()), i=0; 
       tpt1 != NULL; 
       tpt1 = (TilePoint *)(iter.PeekNext()), i++ )
    {
      if( i == uiSelectedPoint )
	{
	  // Insert the new point and iterate past it
	  iter.InsertNext(tptNew, (UINT_32)tptNew);
	  break;
	}
    }

  TesselateOutlineNeighbors(tptNew);

  // New Point Partner
  ASSERT( tpt1 != NULL );

  tptNewPartner = new TilePoint();
  tptNewPartner->SetType(tpOUTLINE);
  tptNewPartner->SetPartner(tptNew);  
  tptNew->SetPartner(tptNewPartner);

  m_pSGParent->FindMatrices(m1, m2, tpt1);
  tptNew->SetPartnerMatrix(m1);
  tptNewPartner->SetPartnerMatrix(m2);
  tptNewPartner->SetPoint(m1*ptL);

  // Insert before tpt1Partner
  tptPartner = tpt1->GetPartner();
  
  iter.Init(&m_listSphereOutlinePoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()), i=0; 
       tpt != NULL; 
       tpt = (TilePoint *)(iter.PeekNext()), i++ )
    {
      if( tpt == tptPartner )
	{
	  if( i == 0 )
	    {
	      m_listSphereOutlinePoints.InsertLast(tptNewPartner, (UINT_32)tptNewPartner);
	    }
	  else
	    {
	      iter.InsertPrev(tptNewPartner, (UINT_32)tptNewPartner);
	    }
	  break;
	}
    }

  tptNew->SetHeightPartner(tptNewPartner);
  tptNewPartner->SetHeightPartner(tptNew);
  TesselateOutlineNeighbors(tptNewPartner);

  m_bSolidModified = TRUE;
}

VOID Tile::DeleteOutlinePoint(UINT_32 uiSelectedPoint)
{
  UINT_32 i;
  CListIterInt iter;
  TilePoint *tpt, *tptPartner, *tptPrev, *tptPost, *tptFirst;

  tpt = NULL;
  tptPartner = NULL;
  tptPrev = NULL;

  iter.Init(&m_listSphereOutlinePoints);  
  tptFirst = (TilePoint *)(iter.PeekFirst());

  for( tpt = (TilePoint *)(iter.PeekFirst()), i=0; 
       tpt != NULL; 
       tpt = (TilePoint *)(iter.PeekNext()), i++ )
    {
      if( i == uiSelectedPoint )
	{
	  if( tpt->GetType() == tpTILE )
	    {
	      SetStatusLabel( "Yellow points cannot be deleted" );
	      return;  // Don't delete tile points
	    }

	  tptPartner = tpt->GetPartner();
	  tptPost = (TilePoint *)(iter.PeekNext());

	  if( tptPost == NULL )
	    {
	      tptPost = tptFirst;
	    }

	  break;
	}

      tptPrev = tpt;
    }

  ASSERT( (tpt != NULL) && (tptPost != NULL) && (tptPrev != NULL) );

  m_listSphereOutlinePoints.RemoveUnordered( (UINT_32)tpt );
  TesselateOutline( tptPrev );

  if( tptPartner != NULL )
    {
      for( tpt = (TilePoint *)(iter.PeekFirst()); 
	   tpt != NULL; 
	   tpt = (TilePoint *)(iter.PeekNext()) )
	{
	  if( tpt == tptPartner )
	    {
	      break;
	    }
	  tptPrev = tpt;
	}
      ASSERT( tpt == tptPartner );

      m_listSphereOutlinePoints.RemoveUnordered( (UINT_32)tptPartner );
      TesselateOutline( tptPrev );
    }

  m_bSolidModified = TRUE;
}

VOID Tile::EnumerateOutlinePoints()
{
  UINT_32 i;
  CListIterInt iter;
  TilePoint *tpt;

  iter.Init(&m_listSphereOutlinePoints);  
  for( tpt = (TilePoint *)(iter.PeekFirst()), i=0; 
       tpt != NULL; 
       tpt = (TilePoint *)(iter.PeekNext()), i++ )
    {
      tpt->SetID(i);
    }
}

TilePoint *Tile::AddInnerPoint(Point ptL)
{
  TilePoint *tpt;
  Vector v;

  // 3D point
  tpt = new TilePoint();
  v.Set(ptL.Get());
  v[W] = 0.0;
  v.Normalize();
  tpt->SetPoint(v[X], v[Y], v[Z]);
  tpt->SetType(tpINNER);

  m_listSphereInnerPoints.InsertLast(tpt, (UINT_32)tpt);

  TesselateInnerBefore(tpt);

  m_bSolidModified = TRUE;

  return tpt;
}

TilePoint *Tile::AddInnerPointCopy(UINT_32 uiSelectedType, UINT_32 uiSelectedPoint)
{
  UINT_32 i;
  CListIterInt iter;
  TilePoint *tpt, *tptToCopy;
  Point pt;
  Vector v;
  
  // Find point to copy
  if( uiSelectedType == SELECTED_OUTLINE_POINT )
    iter.Init(&m_listSphereOutlinePoints);
  else if( uiSelectedType == SELECTED_INNER_POINT )
    iter.Init(&m_listSphereInnerPoints);
  else
    ASSERT(FALSE);

  for( tptToCopy = (TilePoint *)(iter.PeekFirst()), i=0; 
       tptToCopy != NULL;
       tptToCopy = (TilePoint *)(iter.PeekNext()), i++ )
    {
      if( i == uiSelectedPoint )
	{
	  break;
	}
    }
  ASSERT( tptToCopy != NULL );
  pt = tptToCopy->GetPoint();

  // 3D point
  tpt = new TilePoint();
  tpt->SetPoint(pt[X], pt[Y], pt[Z]);
  tpt->SetType(tpINNER);

  m_listSphereInnerPoints.InsertLast(tpt, (UINT_32)tpt);

  TesselateInnerBefore(tpt);

  m_bSolidModified = TRUE;

  return tpt;
}

VOID Tile::MoveInnerPoint(UINT_32 uiSelectedPoint, Point pt)
{
  UINT_32 i;
  CListIterInt iter;
  TilePoint *tpt, *tptPrev, *tptPost;
  Matrix m;

  tptPrev = NULL;
  tptPost = NULL;

  iter.Init(&m_listSphereInnerPoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()), i=0; 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()), i++ )
    {
      if( i == uiSelectedPoint )
	{
	  tpt->SetPoint(pt);
	  TesselateInnerNeighbors(tpt);

	  // unsnap
	  tpt->ClearSnapPoint();

	  break;
	}

      tptPrev = tpt;
    }

  m_bSolidModified = TRUE;
}

VOID Tile::InsertInnerPoint(UINT_32 uiSelectedPoint, Point ptL)
{
  UINT_32 i;
  CListIterInt iter;
  TilePoint *tpt1, *tpt2, *tptNew;
  Matrix m1, m2;

  tpt1 = NULL;
  tpt2 = NULL;

  tptNew = new TilePoint();
  tptNew->SetPoint(ptL);
  tptNew->SetType(tpINNER);

  iter.Init(&m_listSphereInnerPoints);
  for( tpt1 = (TilePoint *)(iter.PeekFirst()), i=0; 
       tpt1 != NULL; 
       tpt1 = (TilePoint *)(iter.PeekNext()), i++ )
    {
      if( i == uiSelectedPoint )
	{
	  // Insert the new point and iterate past it
	  iter.InsertNext(tptNew, (UINT_32)tptNew);
	  (TilePoint *)(iter.PeekNext());

	  tpt2 = (TilePoint *)(iter.PeekNext());

	  break;
	}
    }

  ASSERT( (tpt1 != NULL) && (tpt2 != NULL) );

  TesselateInnerNeighbors(tptNew);

  m_bSolidModified = TRUE;
}

VOID Tile::DeleteInnerPoint(UINT_32 uiSelectedPoint)
{
  UINT_32 i;
  CListIterInt iter;
  TilePoint *tpt, *tptPrev;

  tpt = NULL;
  tptPrev = NULL;

  iter.Init(&m_listSphereInnerPoints);  
  for( tpt = (TilePoint *)(iter.PeekFirst()), i=0; 
       tpt != NULL; 
       tpt = (TilePoint *)(iter.PeekNext()), i++ )
    {
      if( i == uiSelectedPoint )
	{
	  break;
	}

      tptPrev = tpt;
    }

  ASSERT( tpt != NULL);

  m_listSphereInnerPoints.RemoveUnordered( (UINT_32)tpt );

  if( tptPrev != NULL )
    {
      if( tptPrev->GetConnected() )
	{
	  tptPrev->SetConnected( tpt->GetConnected() );
	}

      TesselateInner( tptPrev );
    }

  m_bSolidModified = TRUE;
}

VOID Tile::SnapInnerPointToSelected(UINT_32 uiInnerPointIndex, UINT_32 uiSelectedType,
				    UINT_32 uiSelectedIndex)
{
  UINT_32 i;
  CListIterInt iter;
  TilePoint *tpt, *tptSnapTo;
  SnapPoint *pSnapPoint1, *pSnapPoint2;
  
  // Find point to snap to
  if( uiSelectedType == SELECTED_OUTLINE_POINT )
    iter.Init(&m_listSphereOutlinePoints);
  else if( uiSelectedType == SELECTED_INNER_POINT )
    iter.Init(&m_listSphereInnerPoints);
  else
    ASSERT(FALSE);
  
  for( tptSnapTo = (TilePoint *)(iter.PeekFirst()), i=0; 
       tptSnapTo != NULL;
       tptSnapTo = (TilePoint *)(iter.PeekNext()), i++ )
    {
      if( i == uiSelectedIndex )
	{
	  break;
	}
    }
  ASSERT( tptSnapTo != NULL );

  // Find current inner point
  iter.Init(&m_listSphereInnerPoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()), i=0; 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()), i++ )
    {
      if( i == uiInnerPointIndex )
	{
	  break;
	}
    }

  // Snap points
  tpt->SetPoint(tptSnapTo->GetPoint());
  tpt->SetHeightValue(tptSnapTo->GetHeightValue());
  tpt->SetHeightPoint(tptSnapTo->GetHeightPoint());
  TesselateInnerNeighbors(tpt);

  // Keep track of snap relationship
  pSnapPoint1 = tpt->GetSnapPoint();
  pSnapPoint2 = tptSnapTo->GetSnapPoint();

  if( (pSnapPoint1 == NULL) || (pSnapPoint2 == NULL) )
    {  
      if( pSnapPoint1 != NULL )
	{
	  pSnapPoint1->AddTilePoint(tptSnapTo);  
	  tptSnapTo->SetSnapPoint(pSnapPoint1);
	}
      else if( pSnapPoint2 != NULL )
	{
	  pSnapPoint2->AddTilePoint(tpt);
	  tpt->SetSnapPoint(pSnapPoint2);
	}
      else // both null
	{
	  pSnapPoint1 = new SnapPoint();
	  pSnapPoint1->AddTilePoint(tpt);
	  pSnapPoint1->AddTilePoint(tptSnapTo);  

	  tpt->SetSnapPoint(pSnapPoint1);
	  tptSnapTo->SetSnapPoint(pSnapPoint1);
	}
    }
}

VOID Tile::ClearSnapPoint(UINT_32 uiInnerPointIndex)
{
  UINT_32 i;
  CListIterInt iter;
  TilePoint *tpt;

  iter.Init(&m_listSphereInnerPoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()), i=0; 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()), i++ )
    {
      if( i == uiInnerPointIndex )
	{
	  tpt->ClearSnapPoint();
	  break;
	}
    }
}

VOID Tile::AddHolePoint(Point ptL)
{
  TilePoint *tpt;
  Vector v;

  // 3D point
  tpt = new TilePoint();
  v.Set(ptL.Get());
  v[W] = 0.0;
  v.Normalize();
  tpt->SetPoint(v[X], v[Y], v[Z]);
  tpt->SetType(tpHOLE);

  m_listSphereHolePoints.InsertLast(tpt, (UINT_32)tpt);

  m_bSolidModified = TRUE;
}

VOID Tile::MoveHolePoint(UINT_32 uiSelectedPoint, Point pt)
{
  UINT_32 i;
  CListIterInt iter;
  TilePoint *tpt;

  iter.Init(&m_listSphereHolePoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()), i=0; 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()), i++ )
    {
      if( i == uiSelectedPoint )
	{
	  tpt->SetPoint(pt);
	  break;
	}
    }

  m_bSolidModified = TRUE;
}

VOID Tile::DeleteHolePoint(UINT_32 uiSelectedPoint)
{
  UINT_32 i;
  CListIterInt iter;
  TilePoint *tpt;

  iter.Init(&m_listSphereHolePoints);  
  for( tpt = (TilePoint *)(iter.PeekFirst()), i=0; 
       tpt != NULL; 
       tpt = (TilePoint *)(iter.PeekNext()), i++ )
    {
      if( i == uiSelectedPoint )
	{
	  break;
	}
    }
  ASSERT( tpt != NULL);

  m_listSphereHolePoints.RemoveUnordered( (UINT_32)tpt );

  m_bSolidModified = TRUE;
}

BOOL Tile::GetPointHeightSelected(UINT_32 uiSelectedType, UINT_32 uiSelectedPoint)
{
  UINT_32 i;
  CListIterInt iter;
  TilePoint *tpt;

  if( uiSelectedType == SELECTED_OUTLINE_POINT )
    iter.Init(&m_listSphereOutlinePoints);
  else if( uiSelectedType == SELECTED_INNER_POINT )
    iter.Init(&m_listSphereInnerPoints);
  else
    ASSERT(FALSE);

  for( tpt = (TilePoint *)(iter.PeekFirst()), i=0; 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()), i++ )
    {
      if( i == uiSelectedPoint )
	{
	  return(tpt->GetHeightSelected());
	}
    }
  
  ASSERT(FALSE);
  return FALSE;
}

VOID Tile::SetPointHeightSelected(UINT_32 uiSelectedType, UINT_32 uiSelectedPoint, BOOL b)
{
  UINT_32 i;
  CListIterInt iter;
  TilePoint *tpt;

  if( uiSelectedType == SELECTED_OUTLINE_POINT )
    iter.Init(&m_listSphereOutlinePoints);
  else if( uiSelectedType == SELECTED_INNER_POINT )
    iter.Init(&m_listSphereInnerPoints);
  else
    ASSERT(FALSE);

  for( tpt = (TilePoint *)(iter.PeekFirst()), i=0; 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()), i++ )
    {
      if( i == uiSelectedPoint )
	{
	  tpt->SetHeightSelected( b );	   
	  break;
	}
    }

  m_bBasReliefModified = TRUE;
}

VOID Tile::ToglePointHeightSelected(UINT_32 uiSelectedType, UINT_32 uiSelectedPoint)
{
  UINT_32 i;
  CListIterInt iter;
  TilePoint *tpt;
  BOOL b;

  if( uiSelectedType == SELECTED_OUTLINE_POINT )
    iter.Init(&m_listSphereOutlinePoints);
  else if( uiSelectedType == SELECTED_INNER_POINT )
    iter.Init(&m_listSphereInnerPoints);
  else
    ASSERT(FALSE);

  for( tpt = (TilePoint *)(iter.PeekFirst()), i=0; 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()), i++ )
    {
      if( i == uiSelectedPoint )
	{
	  b = tpt->GetHeightSelected();
	  tpt->SetHeightSelected( !b );
	  break;
	}
    }

  m_bBasReliefModified = TRUE;
}

VOID Tile::DeselectAllHeightPoints()
{
  CListIterInt iter;
  TilePoint *tpt;

  iter.Init(&m_listSphereOutlinePoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()); 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()) )
    {
      tpt->SetHeightSelected(FALSE);
    }

  iter.Init(&m_listSphereInnerPoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()); 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()) )
    {
      tpt->SetHeightSelected(FALSE);
    }

  m_bBasReliefModified = TRUE;
}

VOID Tile::AddHeightToPoints(FLOAT fHeight)
{
  CListIterInt iter;
  TilePoint *tpt, *tptPartner;

  iter.Init(&m_listSphereInnerPoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()); 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()) )
    {
      tpt->SetSnapPointSnapped(FALSE);
    }

  iter.Init(&m_listSphereOutlinePoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()); 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()) )
    {
      tpt->SetSnapPointSnapped(FALSE);
    }

  /////////////////////////////////////////////////

  iter.Init(&m_listSphereInnerPoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()); 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()) )
    {
      if( tpt->GetHeightSelected() )
	{
	  tpt->AddHeight(fHeight);	  
	}
    }

  iter.Init(&m_listSphereOutlinePoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()); 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()) )
    {
      if( tpt->GetHeightSelected() )
	{
	  tpt->AddHeight(fHeight);

	  tptPartner = tpt->GetHeightPartner();

	  if( tptPartner != NULL )
	    {
	      // Don't add height to height partner if already selected
	      if( !(tptPartner->GetHeightSelected()) )
		{
		  tptPartner->AddHeight(fHeight);
		}
	    }	  
	}
    }

  m_bSolidModified = TRUE;
  m_bBasReliefModified = TRUE;
}

Vector Tile::GetPointNormalVector(UINT_32 uiSelectedPoint)
{
  UINT_32 i;
  CListIterInt iter;
  TilePoint *tpt;
  Point pt;
  Vector v;

  iter.Init(&m_listSphereOutlinePoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()), i=0; 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()), i++ )
    {
      if( i == uiSelectedPoint )
	{
	  pt = tpt->GetPoint();
	  v.Set(pt[X], pt[Y], pt[Z]);
	  v.Normalize();
	  return v;
	}
    }

  iter.Init(&m_listSphereInnerPoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()), i=0; 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()), i++ )
    {
      if( i == uiSelectedPoint )
	{
	  pt = tpt->GetPoint();
	  v.Set(pt[X], pt[Y], pt[Z]);
	  v.Normalize();
	  return v;
	}
    }

  ASSERT(FALSE);
  return v;
}

////////////////////////////////////////////////////////////////////////////////

VOID Tile::CalculateNorthPole()
{
  UINT_32 uiPoints;
  CListIterInt iter;
  TilePoint *tpt;
  Point ptCentroid;
  Vector v;

  uiPoints = 0;

  // Edges
  iter.Init(&m_listSphereOutlinePoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()); 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()) )
    {
      tpt->CalculateCentroid(ptCentroid, uiPoints);
    }

  ptCentroid = ptCentroid / (FLOAT)uiPoints;
  v.Set(ptCentroid.Get());
  v[W] = 0.0;
  v.Normalize();

  m_ptNorthPole = -v;
  m_ptNorthPole[W] = 1.0;
}

Point Tile::StereographicProjection(Point pt)
{
  FLOAT t;
  Point ptMapped, pNP;

  ASSERT( pt[Z] != 1.0 );

  pNP.Set(0, 0, 1);           // Using North Pole = (0, 0, 1)
  t = 1.0 / (1.0 - pt[Z]);
  ptMapped = t*pt + (1-t)*pNP;
  ptMapped[W] = 1;

  return ptMapped;
}

VOID Tile::CalculateStereographicPoints()
{
  FLOAT fTheta;
  CListIterInt iter;
  TilePoint *tpt;
  Point pt;
  Vector v, v1, v2;
  Matrix m1, m2;

  CalculateNorthPole();

  v1.Set( 0, 0, 1 );
  v2 = m_ptNorthPole;
  v = v2 * v1;
  fTheta = AngleBetween(v1, v2);

  m1.SetRotate(v, fTheta);
  m2.SetRotate(v, -fTheta);

  // Outer points
  iter.Init(&m_listSphereOutlinePoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()); 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()) )
    {
      pt = tpt->GetPoint();
      pt = m1 * pt;
      pt = StereographicProjection(pt);
      pt = m2 * pt;

      //printf( "Mapped: %.3f %.3f %.3f\n", pt[X], pt[Y], pt[Z]);
      tpt->SetStereographicPoint(pt);
    }

  // Inner points
  iter.Init(&m_listSphereInnerPoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()); 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()) )
    {
      pt = tpt->GetPoint();
      pt = m1 * pt;
      pt = StereographicProjection(pt);
      pt = m2 * pt;

      //printf( "Mapped: %.3f %.3f %.3f\n", pt[X], pt[Y], pt[Z]);
      tpt->SetStereographicPoint(pt);
    }

  // Hole points
  iter.Init(&m_listSphereHolePoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()); 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()) )
    {
      pt = tpt->GetPoint();
      pt = m1 * pt;
      pt = StereographicProjection(pt);
      pt = m2 * pt;

      //printf( "Mapped: %.3f %.3f %.3f\n", pt[X], pt[Y], pt[Z]);
      tpt->SetStereographicPoint(pt);
    }
}

//VOID Tile::Triangulate()
VOID Tile::Foo()
{
  // This produces the correct results

  // Flatten points to 2D and calculate indices
  CalculateStereographicPoints();
  CalculateIndexedPoints();

  ClearMesh(triMesh);
  ClearMesh(in);

  ////////////////////////////////////////////////////////////////////////////
  in.numberofpoints = 14;
  in.numberofpointattributes = 1;
  in.pointlist = (REAL *) malloc(in.numberofpoints * 2 * sizeof(REAL));
  in.pointlist[0] = 0;
  in.pointlist[1] = -2.98858;
  in.pointlist[2] = 4.24264;
  in.pointlist[3] = -0.816497;
  in.pointlist[4] = 0;
  in.pointlist[5] = 0.800789; 
  in.pointlist[6] = -4.24264;
  in.pointlist[7] = -0.816497;
  in.pointlist[8] = -0.670455;
  in.pointlist[9] = -1.35979;
  in.pointlist[10] = 0.261803;
  in.pointlist[11] = -0.862612; 
  in.pointlist[12] = -0.670455;  
  in.pointlist[13] = -1.35979;
  in.pointlist[14] = 0;         
  in.pointlist[15] = -2.98858;
  in.pointlist[16] = -4.24264;   
  in.pointlist[17] = -0.816497; 
  in.pointlist[18] = -0.670455;  
  in.pointlist[19] = -1.35979;
  in.pointlist[20] = 0;          
  in.pointlist[21] = 0.800789;
  in.pointlist[22] = 0.261803;
  in.pointlist[23] = -0.862612;
  in.pointlist[24] = 0.261803;
  in.pointlist[25] = -0.862612;
  in.pointlist[26] = 4.24264;
  in.pointlist[27] = -0.816497;  
  in.pointattributelist = (REAL *) malloc(in.numberofpoints * in.numberofpointattributes *
                                          sizeof(REAL));
  in.pointattributelist[0] = 0;
  in.pointattributelist[1] = 0;
  in.pointattributelist[2] = 0;
  in.pointattributelist[3] = 0;
  in.pointattributelist[4] = 0;
  in.pointattributelist[5] = 0;
  in.pointattributelist[6] = 0;
  in.pointattributelist[7] = 0;
  in.pointattributelist[8] = 0;
  in.pointattributelist[9] = 0;
  in.pointattributelist[10] = 0;
  in.pointattributelist[11] = 0;
  in.pointattributelist[12] = 0;
  in.pointattributelist[13] = 0;
  in.pointattributelist[14] = 0;

  /*
  in.pointmarkerlist = (int *) malloc(in.numberofpoints * sizeof(int));
  in.pointmarkerlist[0] = 0;
  in.pointmarkerlist[1] = 0;
  in.pointmarkerlist[2] = 0;
  in.pointmarkerlist[3] = 0;
  in.pointmarkerlist[4] = 0;
  in.pointmarkerlist[5] = 0;
  in.pointmarkerlist[6] = 0;
  in.pointmarkerlist[7] = 0;
  in.pointmarkerlist[8] = 0;
  in.pointmarkerlist[9] = 0;
  in.pointmarkerlist[10] = 0;
  in.pointmarkerlist[11] = 0;
  in.pointmarkerlist[12] = 0;
  in.pointmarkerlist[13] = 0;
  in.pointmarkerlist[14] = 0;
  */

  in.numberofsegments = 9;
  in.segmentlist = (int *) malloc(in.numberofsegments * 2 * sizeof(int));
  in.segmentlist[0] = 0;
  in.segmentlist[1] = 1;
  in.segmentlist[2] = 1;
  in.segmentlist[3] = 2;
  in.segmentlist[4] = 2;
  in.segmentlist[5] = 3;
  in.segmentlist[6] = 3;
  in.segmentlist[7] = 0;
  in.segmentlist[8] = 4;
  in.segmentlist[9] = 5;
  in.segmentlist[10] = 6;
  in.segmentlist[11] = 7;
  in.segmentlist[12] = 8;
  in.segmentlist[13] = 9;
  in.segmentlist[14] = 10;
  in.segmentlist[15] = 11;
  in.segmentlist[16] = 12;
  in.segmentlist[17] = 13;

  in.numberofholes = 1;
  in.holelist = (REAL *) malloc(in.numberofholes * 2 * sizeof(REAL));
  in.holelist[0] = 0.887978;
  in.holelist[1] = -1.51053;

  in.numberofregions = 0;

  /* ------------------------------------------------------------------ */

  printf("Input point set:\n\n");
  report(&in, 1, 0, 0, 1, 1, 0, 1);

  /* Triangulate the points.  Switches are chosen to read and write a  */
  /*   PSLG (p), preserve the convex hull (c), number everything from  */
  /*   zero (z), assign a regional attribute to each element (A), and  */
  /*   produce an edge list (e), a Voronoi diagram (v), and a triangle */
  /*   neighbor list (n).                                              */

  /* triangulate("pczAevn", &in, &out, &vorout); */
  //triangulate("pze", &in, &triMesh, (struct triangulateio *) NULL);

  printf("Initial triangulation:\n\n");
  report(&triMesh, 1, 1, 0, 1, 1, 0, 1);

  ////////////////////////////////////////////////
  CalculateHoleEdges();
}

Point ScaleStereographicPoint(Point pt)
{
  Point ptScaled;
  FLOAT x, y;

  // Do this so Delaunay triangulation works better

  x = 10.0 * pt[X];
  y = 10.0 * pt[Y];

  if( (-0.0000001 < x) && (x < 0.0000001) )
    x = 0.0;
  
  if( (-0.0000001 < y) && (y < 0.0000001) )
    y = 0.0;
  
  ptScaled[X] = x;
  ptScaled[Y] = y;

  return ptScaled;
}

VOID Tile::ClearMesh(struct triangulateio &mesh)
{
  safefree(mesh.pointlist);
  safefree(mesh.pointattributelist);
  safefree(mesh.pointmarkerlist);
  safefree(mesh.trianglelist);
  safefree(mesh.triangleattributelist);
  safefree(mesh.trianglearealist);
  safefree(mesh.neighborlist);
  safefree(mesh.segmentlist);
  safefree(mesh.segmentmarkerlist);
  safefree(mesh.holelist);
  safefree(mesh.regionlist);
  safefree(mesh.edgelist);
  safefree(mesh.edgemarkerlist);
  safefree(mesh.normlist);

  InitMesh(mesh);
}

VOID Tile::InitMesh(struct triangulateio &mesh)
{
  mesh.pointlist = (REAL *) NULL;
  mesh.pointattributelist = (REAL *) NULL;
  mesh.pointmarkerlist = (int *) NULL;
  mesh.trianglelist = (int *) NULL;
  mesh.triangleattributelist = (REAL *) NULL;
  mesh.trianglearealist = (REAL *) NULL;
  mesh.neighborlist = (int *) NULL;
  mesh.segmentlist = (int *) NULL;
  mesh.segmentmarkerlist = (int *) NULL;
  mesh.holelist = (REAL *) NULL;
  mesh.regionlist = (REAL *) NULL;
  mesh.edgelist = (int *) NULL;
  mesh.edgemarkerlist = (int *) NULL;
  mesh.normlist = (REAL *) NULL;

  mesh.numberofpoints = 0;
  mesh.numberofpointattributes = 0;
  mesh.numberoftriangles = 0;
  mesh.numberofcorners = 0;
  mesh.numberoftriangleattributes = 0;
  mesh.numberofsegments = 0;
  mesh.numberofedges = 0;
  mesh.numberofholes = 0;
  mesh.numberofregions = 0;
}

BOOL Tile::SphereOutlinePoint(TilePoint *tpt)
{
  CListIterInt iter;
  TilePoint *tptOutline;

  iter.Init(&m_listSphereOutlinePoints);
  for( tptOutline = (TilePoint *)(iter.PeekFirst()); 
       tptOutline != NULL;
       tptOutline = (TilePoint *)(iter.PeekNext()) )
    {
      if( tpt == tptOutline )
	{
	  return TRUE;
	}
    }

  return FALSE;
}

VOID Tile::Triangulate()
{
  //TriangulateByAPI();
  TriangulateByFile();
}

VOID Tile::TriangulateByAPI()
{
  CListIterInt iter;
  TilePoint *tpt;
  UINT_32 i, j, uiPoints, uiOuterPoints, uiInnerPoints, uiSegments, uiHolePoints;
  UINT_32 uiPointAttributes, uiSize;
  Point pt;

  // Flatten points to 2D and calculate indices
  CalculateStereographicPoints();
  CalculateIndexedPoints();

  ////////////////////////////////////////////////////////////
  // Free all allocated arrays, including those allocated by Triangle.
  // Initialize in and out

  ClearMesh(triMesh);
  ClearMesh(in);

  ////////////////////////////////////////////////////////////
  // Define points for triangulation
  uiOuterPoints = m_listSphereOutlinePoints.GetLength();
  uiInnerPoints = m_listSphereInnerPoints.GetLength();
  uiHolePoints = m_listSphereHolePoints.GetLength();
  uiPoints = uiOuterPoints + uiInnerPoints;

  in.numberofpoints = uiPoints;
  in.pointlist = (REAL *) malloc(in.numberofpoints * 2 * sizeof(REAL));

  ////////////////////////////////////////////////////////////
  // points (markers and attributes)
  uiSize = in.numberofpoints * 2;

  iter.Init(&m_listSphereOutlinePoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()), i=0; 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()), i++ )
    {
      pt = tpt->GetStereographicPoint();
      pt = ScaleStereographicPoint(pt);

      ASSERT( (2*i) < uiSize );
      in.pointlist[(2*i)]   = pt[X];
      in.pointlist[(2*i)+1] = pt[Y];
    }
  ASSERT( i == uiOuterPoints );

  iter.Init(&m_listSphereInnerPoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()); 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()), i++ )
    {
      pt = tpt->GetStereographicPoint();
      pt = ScaleStereographicPoint(pt);

      ASSERT( (2*i) < uiSize );
      in.pointlist[(2*i)]   = pt[X];
      in.pointlist[(2*i)+1] = pt[Y];
    }
  ASSERT( i == uiPoints );

  ////////////////////////////////////////////////////////////
  // attrributelist and pointmarkerlist
  in.numberofpointattributes = 1; 
  uiPointAttributes = in.numberofpoints * in.numberofpointattributes;
  in.pointattributelist = (REAL *) malloc(in.numberofpoints * in.numberofpointattributes *
                                          sizeof(REAL));
  in.pointmarkerlist = (int *) malloc(in.numberofpoints * sizeof(int));

  uiSize = in.numberofpoints * in.numberofpointattributes;
  for( i = 0; i < uiPointAttributes; i++ )
    {
      ASSERT( i < uiSize );
      in.pointattributelist[i] = 0;
    }
  ASSERT( (int)i == in.numberofpoints * in.numberofpointattributes );

  uiSize = in.numberofpoints;
  for( i = 0; i < (UINT_32)in.numberofpoints; i++ )
    {
      ASSERT( i < uiSize ); 
      in.pointmarkerlist[i] = 0;
    }
  ASSERT( (int)i == in.numberofpoints );

  ////////////////////////////////////////////////////////////
  // segments
  uiSegments = uiOuterPoints;
  iter.Init(&m_listSphereInnerPoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()); 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()) )
    {
      if( tpt->GetConnected() )
	{
	  uiSegments++;
	}
    }
  in.numberofsegments = uiSegments;
  in.segmentlist = (int *) malloc(in.numberofsegments * 2 * sizeof(int));

  // make simple polygon to triangulate (make sure closed or problems)
  uiSize = in.numberofsegments * 2;

  iter.Init(&m_listSphereOutlinePoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()), i=0; 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()), i++ )
    {
      ASSERT( (2*i) + 1 < uiSize );

      in.segmentlist[(2*i)] = i;

      if( i+1 > uiOuterPoints-1 )
	in.segmentlist[(2*i)+1] = 0;
      else
	in.segmentlist[(2*i)+1] = i+1;
    }
  ASSERT( uiOuterPoints == i );

  iter.Init(&m_listSphereInnerPoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()), j=0; 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()), j++ )
    {
      if( tpt->GetConnected() )
	{
	  ASSERT( (2*i) + 1 < uiSize );

	  in.segmentlist[(2*i)]   = uiOuterPoints+j;
	  in.segmentlist[(2*i)+1] = uiOuterPoints+j+1;

	  i++;
	}
    }
  ASSERT( uiSegments == i );

  in.numberofholes = uiHolePoints;

  if( uiHolePoints > 0 )
    {
      in.holelist = (REAL *) malloc(in.numberofholes * 2 * sizeof(REAL));
      uiSize = in.numberofholes * 2;

      iter.Init(&m_listSphereHolePoints);
      for( tpt = (TilePoint *)(iter.PeekFirst()), i=0; 
	   tpt != NULL;
	   tpt = (TilePoint *)(iter.PeekNext()), i++ )
	{
	  pt = tpt->GetStereographicPoint();
	  pt = ScaleStereographicPoint(pt);

	  ASSERT( (2*i) + 1 < uiSize );
	  in.holelist[(2*i)+0] = pt[X]; 
	  in.holelist[(2*i)+1] = pt[Y]; 
	}
      ASSERT( uiHolePoints == i);
    }

  in.numberofregions = 0;  


  printf("Input point set:\n\n");
  report(&in, 1, 0, 0, 1, 1, 0, 1);

  ////////////////////////////////////////////////////////////
  // Triangulate points. Switch -pze (c)
  //triangulate("pze", &in, &triMesh, (struct triangulateio *) NULL);

  printf("Triangulation:\n\n");
  report(&triMesh, 1, 1, 0, 1, 1, 0, 1);

  printf("------------------------------------------------------\n");

  ////////////////////////////////////////////////
  CalculateHoleEdges();
}

VOID Tile::TriangulateByFile()
{
  CListIterInt iter;
  TilePoint *tpt;
  UINT_32 i, j, uiPoints, uiOuterPoints, uiInnerPoints, uiSegments, uiHolePoints;
  Point pt;
  FILE *fpFile;

  // Flatten points to 2D and calculate indices
  CalculateStereographicPoints();
  CalculateIndexedPoints();

  /////////////////////////////////////////////////////////////////////////////////////
  fpFile = fopen( "es_tri.poly", "w");

  uiOuterPoints = m_listSphereOutlinePoints.GetLength();
  uiInnerPoints = m_listSphereInnerPoints.GetLength();
  uiHolePoints = m_listSphereHolePoints.GetLength();
  uiPoints = uiOuterPoints + uiInnerPoints;
  
  // <#points> <dimension (must be 2)> <#attributes> <#boundarymarkers (0 or 1)> 
  fprintf(fpFile, "# vertices, dimension, attributes, boundarymarkers\n" );
  fprintf(fpFile, "%lu 2 0 0\n\n", uiPoints );

  ///////////////////////////////////////////////
  // points

  fprintf(fpFile, "# vertices\n" );

  iter.Init(&m_listSphereOutlinePoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()), i=0; 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()), i++ )
    {
      pt = tpt->GetStereographicPoint();
      pt = ScaleStereographicPoint(pt);

      fprintf(fpFile, "%lu  %f %f\n", i, pt[X], pt[Y] );
    }
  ASSERT( i == uiOuterPoints );

  iter.Init(&m_listSphereInnerPoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()); 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()), i++ )
    {
      pt = tpt->GetStereographicPoint();
      pt = ScaleStereographicPoint(pt);

      fprintf(fpFile,  "%lu  %f %f\n", i, pt[X], pt[Y] );
    }
  ASSERT( i == uiPoints );

  //////////////////////////////////////////////
  // segments

  fprintf(fpFile, "\n# segments\n" );

  uiSegments = uiOuterPoints;
  iter.Init(&m_listSphereInnerPoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()); 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()) )
    {
      if( tpt->GetConnected() )
	{
	  uiSegments++;
	}
    }

  // #segments, #boundarymarkers
  fprintf(fpFile, "%lu 0\n", uiSegments );

  // make simple polygon to triangulate (make sure closed or problems)
  iter.Init(&m_listSphereOutlinePoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()), i=0; 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()), i++ )
    {
      if( i+1 > uiOuterPoints-1 )
	fprintf(fpFile,  "%lu  %lu 0\n", i, i);
      else
	fprintf(fpFile,  "%lu  %lu %lu\n", i, i, i+1);
    }
  ASSERT( uiOuterPoints == i );

  iter.Init(&m_listSphereInnerPoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()), j=0; 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()), j++ )
    {
      if( tpt->GetConnected() )
	{
	  fprintf(fpFile,  "%lu  %lu %lu\n", i, uiOuterPoints+j, uiOuterPoints+j+1 );

	  i++;
	}
    }
  ASSERT( uiSegments == i );

  ///////////////////////////////////////////////
  // holes

  fprintf(fpFile, "\n# holes\n" );
  fprintf(fpFile, "%lu\n", uiHolePoints );

  iter.Init(&m_listSphereHolePoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()), i=0; 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()), i++ )
    {
      pt = tpt->GetStereographicPoint();
      pt = ScaleStereographicPoint(pt);

      fprintf(fpFile, "%lu  %f %f\n", i, pt[X], pt[Y] );
    }
  ASSERT( uiHolePoints == i);

  fprintf(fpFile,  "\n\n" );

  fclose(fpFile);

  
  /////////////////////////////////////////////////////////////////////////////////////
  // Triangulate points. Switch -pze (c)

  if( gbCreateTriFiles )
    {
      //ExecuteTCLCommand( gpcTriangleCommand );
      mainTri(3, gppcTriangleArgs);
    }

  /////////////////////////////////////////////////////////////////////////////////////
  // Read In Results of Triangulation
  UINT_32 uiEdges, uiBoundaryMarkers, uiBoundaryMarker;
  UINT_32 uiID, uiPointIndex1, uiPointIndex2, uiPointIndex3;
  UINT_32 uiTriangles, uiCorners, uiAttributes;
  UINT_32 uiIndex, uiSize;

  ClearMesh(triMesh);

  //////////////////////////////////////////////
  // Hole Edges es_tri.1.edge
  fpFile = fopen( "es_tri.1.edge", "r");
  {
    // <# of edges> <# of boundary markers (0 or 1)> 
    fscanf(fpFile, "%ld %ld", &uiEdges, &uiBoundaryMarkers);

    triMesh.numberofedges = uiEdges;
    triMesh.edgelist = new int[triMesh.numberofedges * 2];
    triMesh.edgemarkerlist = new int[triMesh.numberofedges];

    uiSize = triMesh.numberofedges * 2;
    for( i = 0; i < uiEdges; i++ )
    {
      // <edge #> <endpoint> <endpoint> [boundary marker] 
      fscanf(fpFile, "%ld %ld %ld %ld", &uiID, &uiPointIndex1, &uiPointIndex2, 
	     &uiBoundaryMarker);

      if( i != uiID )
	{
	  printf( "i:%lu   ID: %lu\n", i, uiID);
	  ASSERT( i == uiID );
	}

      uiIndex = 2*i;

      ASSERT( uiIndex + 1 < uiSize );
      triMesh.edgelist[uiIndex] = uiPointIndex1;
      triMesh.edgelist[uiIndex+1] = uiPointIndex2;

      ASSERT( (int)i < triMesh.numberofedges );
      triMesh.edgemarkerlist[i] = uiBoundaryMarker;
    }
  }
  fclose(fpFile);

  //////////////////////////////////////////////
  // Triangles es_tri.1.ele
  fpFile = fopen( "es_tri.1.ele", "r");
  {
    // <# of triangles> <points per triangle> <# of attributes> 
    fscanf(fpFile, "%ld %ld %ld", &uiTriangles, &uiCorners, &uiAttributes);

    ASSERT( uiCorners == 3 );
    triMesh.numberoftriangles = uiTriangles;
    triMesh.numberofcorners = uiCorners;
    triMesh.trianglelist = new int[triMesh.numberoftriangles * triMesh.numberofcorners];

    uiSize = triMesh.numberoftriangles * triMesh.numberofcorners;
    for( i = 0; i < uiTriangles; i++ )
    {
      // <triangle #> <point> <point> <point>
      fscanf(fpFile, "%ld %ld %ld %ld", &uiID,&uiPointIndex1,&uiPointIndex2,&uiPointIndex3);

      ASSERT( i == uiID );

      uiIndex = i * triMesh.numberofcorners;
      ASSERT( uiIndex + 2 < uiSize );
      triMesh.trianglelist[uiIndex + 0] = uiPointIndex1;
      triMesh.trianglelist[uiIndex + 1] = uiPointIndex2;
      triMesh.trianglelist[uiIndex + 2] = uiPointIndex3;
    }
  }
  fclose(fpFile);

  ////////////////////////////////////////////////////////////////
  CalculateHoleEdges();
}

////////////////////////////////////////////////////////////////////////////////

VOID Tile::PrintVertexFacePoints()
{
  CListIterInt iter;
  Point *pt;

  iter.Init(&m_listFaceGridPoints);
  for( pt = (Point *)(iter.PeekFirst()); 
       pt != NULL;
       pt = (Point *)(iter.PeekNext()) )
    {
      glNormal3f(0.0, 0.0, 1.0);
      glVertex3fv(pt->Get());
    }
}

VOID Tile::PrintVertexTilePoints()
{
  CListIterInt iter;
  Point *pt;

  iter.Init(&m_listTileGridPoints);
  for( pt = (Point *)(iter.PeekFirst()); 
       pt != NULL;
       pt = (Point *)(iter.PeekNext()) )
      {
	glVertex3fv(pt->Get());
      }
}

VOID Tile::RenderSphereFace()
{
  CListIterInt iter;
  Point *pt;

  glBegin(GL_LINE_LOOP);
  {
    iter.Init(&m_listSphereFaceGridPoints);
    for( pt = (Point *)(iter.PeekFirst()); 
	 pt != NULL;
	 pt = (Point *)(iter.PeekNext()) )
      {
	glVertex3fv(pt->Get());
      }
  }
  glEnd();
}

VOID Tile::RenderSphereTile()
{
  CListIterInt iter;
  Point *pt;

  glBegin(GL_LINE_LOOP);
  {
    iter.Init(&m_listSphereTileGridPoints);
    for( pt = (Point *)(iter.PeekFirst()); 
	 pt != NULL;
	 pt = (Point *)(iter.PeekNext()) )
      {
	glVertex3fv(pt->Get());
      }
  }
  glEnd();
}

VOID Tile::RenderSphereOutlinePoints()
{
  CListIterInt iter;
  TilePoint *tpt;
  Point pt;

  glMatrixMode(GL_MODELVIEW);
  
  iter.Init(&m_listSphereOutlinePoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()); 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()) )
    {
      if( tpt->GetType() == tpTILE )
	{
	  glColor3f(1.0, 1.0, 0.0); // yellow
	}
      else
	{
	  glColor3f(0.0, 0.0, 1.0); // blue
	}

      pt = tpt->GetPoint();

      glPushMatrix();
      {
	glTranslatef(pt[X], pt[Y], pt[Z]);
	if( !gbZoomIn )
	  glCallList(GLuiSphere02_8);
	else
	  glCallList(GLuiSphere005_8);
      }
      glPopMatrix();
    }
}

VOID Tile::RenderSphereOutlineEdges()
{
  CListIterInt iter;
  TilePoint *tpt;

  // Edges
  glBegin(GL_LINE_LOOP);
  {    
    iter.Init(&m_listSphereOutlinePoints);
    for( tpt = (TilePoint *)(iter.PeekFirst()); 
	 tpt != NULL;
	 tpt = (TilePoint *)(iter.PeekNext()) )
      {
	tpt->PrintVertexTesselatedPoints();
      }
  }
  glEnd();
}

VOID Tile::RenderSphereRaisedOutlineEdges()
{
  CListIterInt iter;
  TilePoint *tpt;

  // Edges
  glBegin(GL_LINE_LOOP);
  {    
    iter.Init(&m_listSphereOutlinePoints);
    for( tpt = (TilePoint *)(iter.PeekFirst()); 
	 tpt != NULL;
	 tpt = (TilePoint *)(iter.PeekNext()) )
      {
	tpt->PrintVertexRaisedTesselatedPoints();
      }
  }
  glEnd();
}

VOID Tile::RenderSphereInnerPoints()
{
  CListIterInt iter;
  TilePoint *tpt;
  Point pt;

  glMatrixMode(GL_MODELVIEW);
  
  iter.Init(&m_listSphereInnerPoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()); 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()) )
    {
      glColor3f(1.0, 0.0, 1.0); // magenta

      pt = tpt->GetPoint();
      
      glPushMatrix();
      {
	glTranslatef(pt[X], pt[Y], pt[Z]);
	if( !gbZoomIn )
	  glCallList(GLuiSphere02_8);
	else
	  glCallList(GLuiSphere005_8);
      }
      glPopMatrix();
    }
}

VOID Tile::RenderSphereInnerEdges()
{
  CListIterInt iter;
  TilePoint *tpt;

  // Edges
  glBegin(GL_LINE_STRIP);
  {    
    iter.Init(&m_listSphereInnerPoints);
    for( tpt = (TilePoint *)(iter.PeekFirst()); 
	 tpt != NULL;
	 tpt = (TilePoint *)(iter.PeekNext()) )
      {
	if( !tpt->GetConnected() )
	  {
	    glVertex3fv((tpt->GetPoint()).Get());

	    glEnd();
	    glBegin(GL_LINE_STRIP);
	  }
	else
	  {
	    tpt->PrintVertexTesselatedPoints();
	  }
      }
  }
  glEnd();
}

VOID Tile::RenderHeightOutlinePoints()
{
  CListIterInt iter;
  TilePoint *tpt;
  Point pt;

  glMatrixMode(GL_MODELVIEW);

  iter.Init(&m_listSphereOutlinePoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()); 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()) )
    {
      pt = tpt->GetHeightPoint();
      
      if( tpt->GetHeightSelected() )
	glColor3f(1.0, 0.5, 0.0); // orange
      else
	glColor3f(0.0, 0.0, 1.0); // blue

      glPushMatrix();
      {
	glTranslatef(pt[X], pt[Y], pt[Z]);
	if( !gbZoomIn )
	  glCallList(GLuiSphere02_8);
	else
	  glCallList(GLuiSphere005_8);
      }
      glPopMatrix();
    }
}

VOID Tile::RenderHeightOutlinePointNormals()
{
  CListIterInt iter;
  TilePoint *tpt;
  Vector v;
  Point pt, p1, p2;

  p1.Set(0, 0, 0);

  glBegin(GL_LINE_STRIP);
  {    
    iter.Init(&m_listSphereOutlinePoints);
    for( tpt = (TilePoint *)(iter.PeekFirst()); 
	 tpt != NULL;
	 tpt = (TilePoint *)(iter.PeekNext()) )
      {
	if( tpt->GetHeightSelected() )
	  {
	    // Draw Normal
	    pt = tpt->GetPoint();

	    v.Set(pt[X], pt[Y], pt[Z]);
	    v.Normalize();
	    v *= 2.0;
	    p2.Set(v[X], v[Y], v[Z]);
	    
	    glColor3f(1.0, 1.0, 1.0); // white
	    
	    glVertex3fv( p1.Get() );
	    glVertex3fv( p2.Get() );
	  }
      }
  }
  glEnd();
}

VOID Tile::RenderHeightOutlineEdges()
{
  CListIterInt iter;
  TilePoint *tpt;

  // Edges
  glBegin(GL_LINE_STRIP);
  {    
    iter.Init(&m_listSphereOutlinePoints);
    for( tpt = (TilePoint *)(iter.PeekFirst()); 
	 tpt != NULL;
	 tpt = (TilePoint *)(iter.PeekNext()) )
      {
	glVertex3fv((tpt->GetHeightPoint()).Get());	
      }

    tpt = (TilePoint *)(iter.PeekFirst());
    if( tpt != NULL )
      {
	glVertex3fv((tpt->GetHeightPoint()).Get());	
      }
  }
  glEnd();
}

VOID Tile::RenderHeightInnerPoints()
{
  CListIterInt iter;
  TilePoint *tpt;
  Point pt;

  glMatrixMode(GL_MODELVIEW);

  iter.Init(&m_listSphereInnerPoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()); 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()) )
    {
      pt = tpt->GetHeightPoint();
      
      if( tpt->GetHeightSelected() )
	glColor3f(1.0, 0.5, 0.0); // orange
      else
	glColor3f(1.0, 0.0, 1.0); // magenta

      glPushMatrix();
      {
	glTranslatef(pt[X], pt[Y], pt[Z]);
	if( !gbZoomIn )
	  glCallList(GLuiSphere02_8);
	else
	  glCallList(GLuiSphere005_8);
      }
      glPopMatrix();
    }
}

VOID Tile::RenderHeightInnerPointNormals()
{
  CListIterInt iter;
  TilePoint *tpt;
  Vector v;
  Point pt, p1, p2;

  p1.Set(0, 0, 0);

  glBegin(GL_LINE_STRIP);
  {    
    iter.Init(&m_listSphereInnerPoints);
    for( tpt = (TilePoint *)(iter.PeekFirst()); 
	 tpt != NULL;
	 tpt = (TilePoint *)(iter.PeekNext()) )
      {
	if( tpt->GetHeightSelected() )
	  {
	    // Draw Normal
	    pt = tpt->GetPoint();

	    v.Set(pt[X], pt[Y], pt[Z]);
	    v.Normalize();
	    v *= 2.0;
	    p2.Set(v[X], v[Y], v[Z]);
	    
	    glColor3f(1.0, 1.0, 1.0); // white
	    
	    glVertex3fv( p1.Get() );
	    glVertex3fv( p2.Get() );
	  }
      }
  }
  glEnd();
}

VOID Tile::RenderHeightInnerEdges()
{
  CListIterInt iter;
  TilePoint *tpt;

  // Edges
  glBegin(GL_LINE_STRIP);
  {    
    iter.Init(&m_listSphereInnerPoints);
    for( tpt = (TilePoint *)(iter.PeekFirst()); 
	 tpt != NULL;
	 tpt = (TilePoint *)(iter.PeekNext()) )
      {
	glVertex3fv((tpt->GetHeightPoint()).Get());
	
	if( !tpt->GetConnected() )
	  {
	    glEnd();
	    glBegin(GL_LINE_STRIP);
	  }
      }
  }
  glEnd();
}

VOID Tile::RenderSphereHolePoints()
{
  CListIterInt iter;
  TilePoint *tpt;
  Point pt;

  glMatrixMode(GL_MODELVIEW);
  
  iter.Init(&m_listSphereHolePoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()); 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()) )
    {
      glColor3f(0.1, 0.1, 0.1); // black

      pt = tpt->GetPoint();
      
      glPushMatrix();
      {
	glTranslatef(pt[X], pt[Y], pt[Z]);
	if( !gbZoomIn )
	  glCallList(GLuiSphere02_8);
	else
	  glCallList(GLuiSphere005_8);
      }
      glPopMatrix();
    }
}

VOID Tile::RenderSelectionSphereOutlinePoints()
{
  UINT_32 i;
  CListIterInt iter;
  TilePoint *tpt;
  Point pt;

  glMatrixMode(GL_MODELVIEW);
  
  iter.Init(&m_listSphereOutlinePoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()), i=0; 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()), i++ )
    {
      //if( tpt->GetType() != tpTILE )

      pt = tpt->GetPoint();
      
      glPushMatrix();
      {
	glTranslatef(pt[X], pt[Y], pt[Z]);
	
	glPushName(i);
	{
	  if( !gbZoomIn )
	    glCallList(GLuiSphere02_8);
	  else
	    glCallList(GLuiSphere005_8);
	}
	glPopName();
      }
      glPopMatrix();
    }
}

VOID Tile::RenderSelectionSphereOutlineEdges()
{
  UINT_32 i;
  CListIterInt iter;
  TilePoint *tpt;

  iter.Init(&m_listSphereOutlinePoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()), i=0; 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()), i++ )
    {
      glPushName(i);
      {
	glBegin(GL_LINE_STRIP);
	{	    
	  tpt->PrintVertexTesselatedPoints();
	}
	glEnd();	
      }
      glPopName();
    }
}

VOID Tile::RenderSelectionSphereInnerPoints()
{
  UINT_32 i;
  CListIterInt iter;
  TilePoint *tpt;
  Point pt;

  glMatrixMode(GL_MODELVIEW);
  
  iter.Init(&m_listSphereInnerPoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()), i=0; 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()), i++ )
    {
      if( i != guiCurrentSelectedInner ) // don't draw current inner point being moved
	{
	  pt = tpt->GetPoint();
	  
	  glPushMatrix();
	  {
	    glTranslatef(pt[X], pt[Y], pt[Z]);
	    
	    glPushName(i);
	    {
	      if( !gbZoomIn )
		glCallList(GLuiSphere02_8);
	      else
		glCallList(GLuiSphere005_8);
	    }
	    glPopName();
	  }
	  glPopMatrix();
	}
    }
}

VOID Tile::RenderSelectionSphereInnerEdges()
{
  UINT_32 i;
  CListIterInt iter;
  TilePoint *tpt;

  iter.Init(&m_listSphereInnerPoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()), i=0; 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()), i++ )
    {
      if( tpt->GetConnected() )
	{
	  glPushName(i);
	  {
	    glBegin(GL_LINE_STRIP);
	    {	    
	      tpt->PrintVertexTesselatedPoints();
	    }
	    glEnd();	
	  }
	  glPopName();
	}
    }
}

VOID Tile::RenderSelectionSphereHeightOutlinePoints()
{
  UINT_32 i;
  CListIterInt iter;
  TilePoint *tpt;
  Point pt;

  glMatrixMode(GL_MODELVIEW);
  
  iter.Init(&m_listSphereOutlinePoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()), i=0; 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()), i++ )
    {
      pt = tpt->GetHeightPoint();
      
      glPushMatrix();
      {
	glTranslatef(pt[X], pt[Y], pt[Z]);
	
	glPushName(i);
	{
	  if( !gbZoomIn )
	    glCallList(GLuiSphere02_8);
	  else
	    glCallList(GLuiSphere005_8);
	}
	glPopName();
      }
      glPopMatrix();
    }
}

VOID Tile::RenderSelectionSphereHeightInnerPoints()
{
  UINT_32 i;
  CListIterInt iter;
  TilePoint *tpt;
  Point pt;

  glMatrixMode(GL_MODELVIEW);
  
  iter.Init(&m_listSphereInnerPoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()), i=0; 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()), i++ )
    {
      pt = tpt->GetHeightPoint();
      
      glPushMatrix();
      {
	glTranslatef(pt[X], pt[Y], pt[Z]);
	
	glPushName(i);
	{
	  if( !gbZoomIn )
	    glCallList(GLuiSphere02_8);
	  else
	    glCallList(GLuiSphere005_8);
	}
	glPopName();
      }
      glPopMatrix();
    }
}

VOID Tile::RenderSelectionSphereHolePoints()
{
  UINT_32 i;
  CListIterInt iter;
  TilePoint *tpt;
  Point pt;

  glMatrixMode(GL_MODELVIEW);
  
  iter.Init(&m_listSphereHolePoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()), i=0; 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()), i++ )
    {
      pt = tpt->GetPoint();
      
      glPushMatrix();
      {
	glTranslatef(pt[X], pt[Y], pt[Z]);
	
	glPushName(i);
	{
	  if( !gbZoomIn )
	    glCallList(GLuiSphere02_8);
	  else
	    glCallList(GLuiSphere005_8);
	}
	glPopName();
      }
      glPopMatrix();
    }
}

VOID Tile::RenderMesh()
{
  UINT_32 i, j, index;
  TilePoint *tpt;
  
  glBegin(GL_LINES);
  {
    for (i = 0; i < (UINT_32)triMesh.numberofedges; i++) 
      {
	for (j = 0; j < 2; j++) 
	  {
	    index = triMesh.edgelist[2*i + j];
	    tpt = GetIndexedPoint(index);

	    if( tpt != NULL )
	      glVertex3fv( (tpt->GetStereographicPoint()).Get() );
	  }
      }
  }
  glEnd();
}

VOID Tile::RenderStereographicPoints()
{
  CListIterInt iter;
  TilePoint *tpt;
  Point pt;

  glMatrixMode(GL_MODELVIEW);
  
  glColor3f(0.0, 1.0, 0.0); // Green
  iter.Init(&m_listSphereOutlinePoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()); 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()) )
    {
      pt.Set((tpt->GetStereographicPoint()).Get());

      glPushMatrix();
      {
	glTranslatef(pt[X], pt[Y], pt[Z]);
	if( !gbZoomIn )
	  glCallList(GLuiSphere01_8);
	else
	  glCallList(GLuiSphere005_8);
      }
      glPopMatrix();
    }

  glColor3f(1.0, 0.0, 1.0); // Magenta
  iter.Init(&m_listSphereInnerPoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()); 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()) )
    {
      pt.Set((tpt->GetStereographicPoint()).Get());

      glPushMatrix();
      {
	glTranslatef(pt[X], pt[Y], pt[Z]);	
	if( !gbZoomIn )
	  glCallList(GLuiSphere01_8);
	else
	  glCallList(GLuiSphere005_8);
      }
      glPopMatrix();
    }

  glColor3f(0.3, 0.3, 0.3);  // Black    
  iter.Init(&m_listSphereHolePoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()); 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()) )
    {
      pt.Set((tpt->GetStereographicPoint()).Get());

      glPushMatrix();
      {
	glTranslatef(pt[X], pt[Y], pt[Z]);	
	if( !gbZoomIn )
	  glCallList(GLuiSphere01_8);
	else
	  glCallList(GLuiSphere005_8);
      }
      glPopMatrix();
    }
}

VOID Tile::RenderStereographicEdges()
{
  CListIterInt iter;
  TilePoint *tpt;

  glBegin(GL_LINE_LOOP);
  {	    
    iter.Init(&m_listSphereOutlinePoints);
    for( tpt = (TilePoint *)(iter.PeekFirst()); 
	 tpt != NULL;
	 tpt = (TilePoint *)(iter.PeekNext()) )
      {
	glVertex3fv((tpt->GetStereographicPoint()).Get());
      }
  }
  glEnd();	
}

VOID Tile::RenderMeshTilePoints()
{
  UINT_32 i, j, index;
  TilePoint *tpt;
  Point p1, p2, p3;
  Vector v;

  for (i = 0; i < (UINT_32)triMesh.numberoftriangles; i++) 
    {
      ASSERT( triMesh.numberofcorners >= 3 );
      index = triMesh.trianglelist[i * triMesh.numberofcorners + 0];
      tpt = GetIndexedPoint(index);
      p1 = tpt->GetHeightPoint();
      
      index = triMesh.trianglelist[i * triMesh.numberofcorners + 1];
      tpt = GetIndexedPoint(index);
      p2 = tpt->GetHeightPoint();
      
      index = triMesh.trianglelist[i * triMesh.numberofcorners + 2];
      tpt = GetIndexedPoint(index);
      p3 = tpt->GetHeightPoint();
      
      v = (p3-p2)*(p1-p2);
      v.Normalize();
      
      glNormal3fv( v.Get() );
      
      glBegin(GL_POLYGON);
      {
	for (j = 0; j < (UINT_32)triMesh.numberofcorners; j++) 
	  {
	    index = triMesh.trianglelist[i * triMesh.numberofcorners + j];
	    tpt = GetIndexedPoint(index);
	    
	    if( tpt != NULL )
	      glVertex3fv( (tpt->GetHeightPoint()).Get() );	    
	  }
      }
      glEnd();
    }
}

VOID Tile::CalculateExtrudedPoints(Point &p1, Point &pp1, TilePoint *tpt)
{
  ASSERT( tpt != NULL );

  FLOAT f;

  if( gbExtrudeInward )
    {
      f = (1.0 - m_fThickness);

      if( gbTileDetailed )
	{
	  p1 = tpt->GetHeightPoint();
	  pp1 = f * p1;
	}
      else
	{
	  p1 = tpt->GetHeightPoint();
	  pp1 = f * tpt->GetPoint();
	}
    }
  else
    {
      f = (1.0 + m_fThickness);

      if( gbTileDetailed )
	{
	  pp1 = tpt->GetHeightPoint();
	  p1 = f * pp1;
	}
      else
	{
	  p1 = f * tpt->GetHeightPoint();
	  pp1 = tpt->GetPoint();
	}
    }
}

VOID Tile::CalculateOutputExtrudedPoints(Point &p1, Point &pp1, TilePoint *tpt)
{
  ASSERT( tpt != NULL );

  CalculateExtrudedPoints(p1, pp1, tpt);
  
  p1 *= m_fRadius;
  pp1 *= m_fRadius;
}

VOID Tile::RenderMeshSide(Point pt1, Point pt2, Point pp1, Point pp2)
{
  Vector v;
  
  v = (pt2-pp1)*(pt1-pp1);
  v.Normalize();
  glNormal3fv( v.Get() );      
  glBegin(GL_POLYGON);
  {
    glVertex3fv( pt2.Get() );
    glVertex3fv( pt1.Get() );
    glVertex3fv( pp1.Get() );
  }
  glEnd();
  
  v = (pt2-pp2)*(pp1-pp2);
  v.Normalize();
  glNormal3fv( v.Get() );      
  glBegin(GL_POLYGON);
  {
    glVertex3fv( pt2.Get() );
    glVertex3fv( pp1.Get() );
    glVertex3fv( pp2.Get() );
  }
  glEnd();
}
  
VOID Tile::RenderMeshExtrudedTilePoints(BOOL bBorder)
{
  UINT_32 i, index;
  CListIterInt iter;
  TilePoint *tpt1, *tpt2, *tpt;
  Point p1, p2, p3, pp1, pp2, pp3;
  Vector v;
  BOOL bOutlinePoint;

  glBegin(GL_TRIANGLES);
  {
    for (i = 0; i < (UINT_32)triMesh.numberoftriangles; i++) 
      {
	ASSERT( triMesh.numberofcorners == 3 );
	
	index = triMesh.trianglelist[i * triMesh.numberofcorners + 0];
	tpt = GetIndexedPoint(index);
	CalculateExtrudedPoints(p1, pp1, tpt);
	
	index = triMesh.trianglelist[i * triMesh.numberofcorners + 1];
	tpt = GetIndexedPoint(index);
	CalculateExtrudedPoints(p2, pp2, tpt);
	
	index = triMesh.trianglelist[i * triMesh.numberofcorners + 2];
	tpt = GetIndexedPoint(index);
	CalculateExtrudedPoints(p3, pp3, tpt);
	
	// Top 
	v = (p3-p2)*(p1-p2);
	v.Normalize();

	glNormal3fv( v.Get() );
	glVertex3fv( p1.Get() );	    
	glVertex3fv( p2.Get() );	    
	glVertex3fv( p3.Get() );	    

	// Bottom 
	v = -(pp3-pp2)*(pp1-pp2);
	v.Normalize();

	glNormal3fv( v.Get() );      
	glVertex3fv( pp3.Get() );
	glVertex3fv( pp2.Get() );
	glVertex3fv( pp1.Get() );
      }
  }
  glEnd();
  
  // Boundary - Border and Holes
  if( triMesh.edgemarkerlist != NULL )
    {
      for( i = 0; i < (UINT_32)triMesh.numberofedges; i++ )
	{
	  if( 1 == triMesh.edgemarkerlist[i] )             // Boundary
	    {
	      index = triMesh.edgelist[2*i];
	      tpt1 = GetIndexedPoint(index);
		  
	      bOutlinePoint = SphereOutlinePoint(tpt1);

	      if( !bOutlinePoint || (bOutlinePoint && bBorder) ) // Holes or border
		{
		  index = triMesh.edgelist[2*i+1];
		  tpt2 = GetIndexedPoint(index);
		  
		  CalculateExtrudedPoints(p1, pp1, tpt1);
		  CalculateExtrudedPoints(p2, pp2, tpt2);
		  
		  RenderMeshSide(p1, p2, pp1, pp2);
		}
	    }
	}
    }
}

//////////////////////////////////////////////////////////////////////////////////

VOID Tile::OutputSTL(FILE *fpOutputFile, Matrix mTransform, BOOL bBorder)
{
  ASSERT( fpOutputFile != NULL );

  UINT_32 i, index;
  CListIterInt iter;
  TilePoint *tpt1, *tpt2, *tpt;
  Point p1, p2, p3, pp1, pp2, pp3;
  Vector v;
  BOOL bOutlinePoint;

  for (i = 0; i < (UINT_32)triMesh.numberoftriangles; i++) 
    {
      ASSERT( triMesh.numberofcorners == 3 );

      index = triMesh.trianglelist[i * triMesh.numberofcorners + 0];
      tpt = GetIndexedPoint(index);
      CalculateOutputExtrudedPoints(p1, pp1, tpt);

      index = triMesh.trianglelist[i * triMesh.numberofcorners + 1];
      tpt = GetIndexedPoint(index);
      CalculateOutputExtrudedPoints(p2, pp2, tpt);

      index = triMesh.trianglelist[i * triMesh.numberofcorners + 2];
      tpt = GetIndexedPoint(index);
      CalculateOutputExtrudedPoints(p3, pp3, tpt);

      p1 = mTransform * p1;
      p2 = mTransform * p2;
      p3 = mTransform * p3;
      pp1 = mTransform * pp1;
      pp2 = mTransform * pp2;
      pp3 = mTransform * pp3;

      // Top 
      WriteSTLTriangle(fpOutputFile, p1, p2, p3);

      // Bottom 
      WriteSTLTriangle(fpOutputFile, pp3, pp2, pp1);
    }

  // Boundary - Border and Holes
  if( triMesh.edgemarkerlist != NULL )
    {
      for( i = 0; i < (UINT_32)triMesh.numberofedges; i++ )
	{
	  if( 1 == triMesh.edgemarkerlist[i] )             // Boundary
	    {
	      index = triMesh.edgelist[2*i];
	      tpt1 = GetIndexedPoint(index);
		  
	      bOutlinePoint = SphereOutlinePoint(tpt1);

	      if( !bOutlinePoint || (bOutlinePoint && bBorder) ) // Holes or border
		{
		  index = triMesh.edgelist[2*i+1];
		  tpt2 = GetIndexedPoint(index);
		  
		  CalculateOutputExtrudedPoints(p1, pp1, tpt1);
		  CalculateOutputExtrudedPoints(p2, pp2, tpt2);
		  
		  p1 = mTransform * p1;
		  p2 = mTransform * p2;
		  pp1 = mTransform * pp1;
		  pp2 = mTransform * pp2;
		  
		  WriteSTLTriangle(fpOutputFile, p2, p1, pp1);
		  WriteSTLTriangle(fpOutputFile, p2, pp1, pp2);
		}	  
	    }
	}
    }
}

/////////////////////////////////////////////////////////////////////////////////////////

VOID WriteIndent(FILE *fpOutputFile, UINT_32 uiIndentLevel)
{
  UINT_32 i;

  for ( i = 0; i < uiIndentLevel; i++ )
    {
      fprintf(fpOutputFile, "  ");
    }
}

VOID WriteColor(FILE *fpOutputFile, Vector vColor)
{
  UINT_32 uiIndentLevel;
  
  uiIndentLevel = 1;

  WriteIndent(fpOutputFile, uiIndentLevel);
  fprintf(fpOutputFile, "appearance Appearance {\n");
  WriteIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "material Material {\n");

  WriteIndent(fpOutputFile, uiIndentLevel+2);
  fprintf(fpOutputFile, "diffuseColor %.6f %.6f %.6f\n", 
	  vColor[X], vColor[Y], vColor[Z]);

  WriteIndent(fpOutputFile, uiIndentLevel+2);
  fprintf(fpOutputFile, "ambientIntensity 1\n");
  WriteIndent(fpOutputFile, uiIndentLevel+2);
  fprintf(fpOutputFile, "specularColor 0 0 0\n");
  WriteIndent(fpOutputFile, uiIndentLevel+2);
  fprintf(fpOutputFile, "emissiveColor 0 0 0\n");
  WriteIndent(fpOutputFile, uiIndentLevel+2);
  fprintf(fpOutputFile, "shininess 0\n");
  WriteIndent(fpOutputFile, uiIndentLevel+2);
  fprintf(fpOutputFile, "transparency 0\n");

  WriteIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "} # end material\n");
  WriteIndent(fpOutputFile, uiIndentLevel);
  fprintf(fpOutputFile, "} # end appearance\n\n");
}

VOID Tile::OutputVRML(FILE *fpOutputFile, Matrix mTransform, Vector vColor, BOOL bBorder)
{
  ASSERT( fpOutputFile != NULL );

  UINT_32 i, j, index, uiIndentLevel, uiTriangles, uiVertices;
  UINT_32 uip1, uip2, uipp1, uipp2;
  CListIterInt iter;
  TilePoint *tpt;
  Point p1, pp1;
  BOOL bOutlinePoint;

  fprintf(fpOutputFile, "Shape { # triangle mesh: SPHERE\n\n");

  // Color  
  WriteColor(fpOutputFile, vColor);

  // Mesh
  uiIndentLevel = 1;
  uiVertices = m_uiPoints;
  uiTriangles = (UINT_32)((2 * triMesh.numberoftriangles) + (2 * m_uiHoleEdges));
  if( bBorder )
    {
      uiTriangles += (UINT_32)(2 * m_listSphereOutlinePoints.GetLength());
    }

  // Header
  WriteIndent(fpOutputFile, uiIndentLevel);
  fprintf(fpOutputFile, "geometry IndexedFaceSet {\n");
  WriteIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "ccw    TRUE\n");
  WriteIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "convex TRUE\n");
  WriteIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "solid  FALSE\n\n");

  // Triangles
  WriteIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "# %lu triangles\n", uiTriangles);
  WriteIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "coordIndex [");

  // top
  for (i = 0; i < (UINT_32)triMesh.numberoftriangles; i++) 
    {
      for (j = 0; j < (UINT_32)triMesh.numberofcorners; j++) 
	{
	  index = triMesh.trianglelist[i * triMesh.numberofcorners + j];
	  tpt = GetIndexedPoint(index);
	  
	  if( tpt != NULL )
	    fprintf(fpOutputFile, "%lu,", index);
	}
      fprintf(fpOutputFile, "-1, ");
    }

  // bottom
  for (i = 0; i < (UINT_32)triMesh.numberoftriangles; i++) 
    {
      for (j = (UINT_32)triMesh.numberofcorners; j > 0; j--) 
	{
	  index = triMesh.trianglelist[i * triMesh.numberofcorners + j - 1];
	  tpt = GetIndexedPoint(index);
	  
	  if( tpt != NULL )
	    fprintf(fpOutputFile, "%lu,", uiVertices + index);
	}
      fprintf(fpOutputFile, "-1, ");
    }

  // Boundary - Border and Holes
  if( triMesh.edgemarkerlist != NULL )
    {
      for( i = 0; i < (UINT_32)triMesh.numberofedges; i++ )
	{
	  if( 1 == triMesh.edgemarkerlist[i] ) // Boundary Edge
	    {
	      index = triMesh.edgelist[2*i];
	      tpt = GetIndexedPoint(index);
	      uip1 = index;

	      bOutlinePoint = SphereOutlinePoint(tpt);

	      if( !bOutlinePoint || (bOutlinePoint && bBorder) ) // Holes or border
		{
		  index = triMesh.edgelist[2*i+1];
		  uip2 = index;

		  uipp1 = uiVertices + uip1;		  
		  uipp2 = uiVertices + uip2;

		  fprintf(fpOutputFile, "%lu,%lu,%lu,-1, ", uip2, uip1, uipp1);
		  fprintf(fpOutputFile, "%lu,%lu,%lu,-1, ", uip2, uipp1, uipp2);
		}
	    }	  
	}
    }

  // Points
  fprintf(fpOutputFile, "] \n\n");
  WriteIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "# %lu points\n", 2*uiVertices);
  WriteIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "coord Coordinate {\n");
  WriteIndent(fpOutputFile, uiIndentLevel+2);
  fprintf(fpOutputFile, "point [");

  // top
  for (i = 0; i < uiVertices; i++) 
    {
      tpt = GetIndexedPoint(i);

      CalculateOutputExtrudedPoints(p1, pp1, tpt);
      p1 = mTransform * p1;      
      fprintf(fpOutputFile, "%.6f %.6f %.6f, ", p1[X], p1[Y], p1[Z]);
    }

  // bottom 
  for (i = 0; i < uiVertices; i++) 
    {
      tpt = GetIndexedPoint(i);

      CalculateOutputExtrudedPoints(p1, pp1, tpt);
      pp1 = mTransform * pp1;
      fprintf(fpOutputFile, "%.6f %.6f %.6f, ", pp1[X], pp1[Y], pp1[Z]);
    }

  fprintf(fpOutputFile, "]\n");
  WriteIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "}\n\n");

  /*
  // Normals
  WriteIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "normal Normal {\n");
  WriteIndent(fpOutputFile, uiIndentLevel+2);
  fprintf(fpOutputFile, "vector [");

  //
  // output normals


  fprintf(fpOutputFile, "]\n");
  WriteIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "}\n\n");

  WriteIndent(fpOutputFile, uiIndentLevel+1);
  fprintf(fpOutputFile, "normalPerVertex FALSE\n\n");
  */

  // End
  WriteIndent(fpOutputFile, uiIndentLevel);
  fprintf(fpOutputFile, "} # end geometry\n");

  fprintf(fpOutputFile, "} # end triangle mesh\n\n");
}

/////////////////////////////////////////////////////////////////////////////////////////

VOID Tile::OutputES(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  UINT_32 uiOutlinePoints, uiInnerPoints, uiHolePoints;
  CListIterInt iter;
  TilePoint *tpt;

  uiOutlinePoints = m_listSphereOutlinePoints.GetLength();
  uiInnerPoints = m_listSphereInnerPoints.GetLength();
  uiHolePoints = m_listSphereHolePoints.GetLength();

  fprintf(fpOutputFile, "radius %f  thickness %f extrude %ld shell %ld\n\n", 
	  m_fRadius, m_fThickness, (UINT_32)gbExtrudeInward, (UINT_32)gbTileDetailed );

  // Outline Points 
  fprintf(fpOutputFile, "sphereoutlinepoints %ld\n", uiOutlinePoints);
  iter.Init(&m_listSphereOutlinePoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()); 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()) )
    {
      tpt->OutputES(fpOutputFile);
      fprintf(fpOutputFile, "\n");
    }
  fprintf(fpOutputFile, "\n");

  // Outline Point Partners
  OutputESTilePointPartners(fpOutputFile);
  fprintf(fpOutputFile, "\n");

  // Inner Points
  fprintf(fpOutputFile, "sphereinnerpoints %ld\n", uiInnerPoints);
  iter.Init(&m_listSphereInnerPoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()); 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()) )
    {
      tpt->OutputES(fpOutputFile);
      fprintf(fpOutputFile, "\n");
    }
  fprintf(fpOutputFile, "\n");

  // Hole Points
  fprintf(fpOutputFile, "sphereholepoints %ld\n", uiHolePoints);
  iter.Init(&m_listSphereHolePoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()); 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()) )
    {
      tpt->OutputES(fpOutputFile);
      fprintf(fpOutputFile, "\n");
    }
  fprintf(fpOutputFile, "\n");
}

VOID Tile::OutputESTilePointPartners(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  UINT_32 uiID, uiIDPartner;
  CListIterInt iter, iter2;
  TilePoint *tpt, *tptPartner, *tpt2;

  iter.Init(&m_listSphereOutlinePoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()), uiID=1; 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()), uiID++ )
    {
      tptPartner = tpt->GetPartner();

      iter2.Init(&m_listSphereOutlinePoints);
      for( tpt2 = (TilePoint *)(iter2.PeekFirst()), uiIDPartner=1; 
	   tpt2 != NULL;
	   tpt2 = (TilePoint *)(iter2.PeekNext()), uiIDPartner++)
	{
	  if( tpt2 == tptPartner )
	    {
	      fprintf(fpOutputFile, "partners %ld %ld\n", uiID, uiIDPartner);
	      break;
	    }
	}
    }
}

VOID Tile::ClearSphereOutlinePoints()
{
  m_listSphereOutlinePoints.RemoveAll();
}

VOID Tile::ClearSphereInnerPoints()
{
  m_listSphereInnerPoints.RemoveAll();
}

VOID Tile::ClearSphereHolePoints()
{
  m_listSphereHolePoints.RemoveAll();
}

VOID Tile::CreateTilePoint(Point pt, UINT_32 uiType, UINT_32 uiTilePoint, 
			   UINT_32 uiConnected, FLOAT fHeight, Matrix m)
{
  TilePoint *tpt;

  if( (uiType == tpOUTLINE) || (uiType == tpTILE) )
    {
      tpt = new TilePoint();
      tpt->SetPoint(pt);
      tpt->SetType((TilePointType)uiType);
      tpt->SetTilePoint(uiTilePoint);
      tpt->SetPartnerMatrix(m);

      m_listSphereOutlinePoints.InsertLast(tpt, (UINT_32)tpt);
    }
  else if( uiType == tpINNER )
    {
      tpt = new TilePoint();
      tpt->SetPoint(pt);
      tpt->SetType(tpINNER);
      tpt->SetConnected(uiConnected);
      tpt->SetHeight(fHeight);

      m_listSphereInnerPoints.InsertLast(tpt, (UINT_32)tpt);
    }
  else if( uiType == tpHOLE )
    {
      tpt = new TilePoint();
      tpt->SetPoint(pt);
      tpt->SetType(tpHOLE);

      m_listSphereHolePoints.InsertLast(tpt, (UINT_32)tpt);
    }
}

VOID Tile::SetPartners(UINT_32 uiID1, UINT_32 uiID2)
{
  UINT_32 uiID;
  CListIterInt iter;
  TilePoint *tpt, *tpt1, *tpt2;

  tpt1 = NULL;
  tpt2 = NULL;

  iter.Init(&m_listSphereOutlinePoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()), uiID=1; 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()), uiID++ )
    {
      if( uiID == uiID1 )
	{
	  tpt1 = tpt;
	}

      if( uiID == uiID2 )
	{
	  tpt2 = tpt;
	}

      if( (tpt1 != NULL) && (tpt2 != NULL) )
	{
	  break;
	}
    }

  ASSERT( (tpt1 != NULL) && (tpt2 != NULL ) );

  tpt1->SetPartner(tpt2);
  tpt2->SetPartner(tpt1);
}

VOID Tile::SetHeightPartnersFromOpenES(UINT_32 uiID1, UINT_32 uiID2)
{
  UINT_32 uiID;
  CListIterInt iter;
  TilePoint *tpt, *tpt1, *tpt2;

  tpt1 = NULL;
  tpt2 = NULL;

  iter.Init(&m_listSphereOutlinePoints);
  for( tpt = (TilePoint *)(iter.PeekFirst()), uiID=1; 
       tpt != NULL;
       tpt = (TilePoint *)(iter.PeekNext()), uiID++ )
    {
      if( uiID == uiID1 )
	{
	  tpt1 = tpt;
	}

      if( uiID == uiID2 )
	{
	  tpt2 = tpt;
	}

      if( (tpt1 != NULL) && (tpt2 != NULL) )
	{
	  break;
	}
    }

  ASSERT( (tpt1 != NULL) && (tpt2 != NULL ) );

  tpt1->SetHeightPartner(tpt2);
  tpt2->SetHeightPartner(tpt1);
}
